##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Param,
                           Set,
                           Var,
                           TerminationCondition,
                           SolverStatus,
                           value)
from pyomo.network import Port
from pyomo.common.config import ConfigBlock

from idaes.core import (FlowsheetBlock,
                        StateBlock,
                        declare_process_block_class)
from idaes.property_models.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.property_models import iapws95
from idaes.property_models.examples.saponification_thermo import \
    SaponificationParameterBlock

from idaes.unit_models.mixer import (Mixer,
                                     MixerData,
                                     MixingType,
                                     MaterialBalanceType,
                                     MomentumMixingType)
from idaes.core.util.exceptions import (BurntToast,
                                        ConfigurationError,
                                        PropertyNotSupportedError)
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock,
                                     TestStateBlock)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
# Unit Tests for Mixer
class TestMixer(object):
    @declare_process_block_class("MixerFrame")
    class MixerFrameData(MixerData):
        def build(self):
            super(MixerData, self).build()

    @pytest.fixture(scope="function")
    def mixer_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.pp = PhysicalParameterTestBlock()

        m.fs.mix = MixerFrame(default={"property_package": m.fs.pp})

        return m

    def test_mixer_config(self, mixer_frame):
        assert len(mixer_frame.fs.mix.config) == 12
        assert mixer_frame.fs.mix.config.dynamic is False
        assert mixer_frame.fs.mix.config.has_holdup is False
        assert mixer_frame.fs.mix.config.property_package == mixer_frame.fs.pp
        assert isinstance(mixer_frame.fs.mix.config.property_package_args,
                          ConfigBlock)
        assert len(mixer_frame.fs.mix.config.property_package_args) == 0
        assert mixer_frame.fs.mix.config.inlet_list is None
        assert mixer_frame.fs.mix.config.num_inlets is None
        assert mixer_frame.fs.mix.config.has_phase_equilibrium is False
        assert mixer_frame.fs.mix.config.energy_mixing_type == \
            MixingType.extensive
        assert mixer_frame.fs.mix.config.momentum_mixing_type == \
            MomentumMixingType.minimize
        assert mixer_frame.fs.mix.config.mixed_state_block is None
        assert mixer_frame.fs.mix.config.construct_ports is True
        assert mixer_frame.fs.mix.config.material_balance_type == \
            MaterialBalanceType.useDefault

    def test_inherited_methods(self, mixer_frame):
        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        assert hasattr(mixer_frame.fs.mix.config.property_package,
                       "phase_list")

    def test_create_inlet_list_default(self, mixer_frame):
        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()

        for i in inlet_list:
            assert i in ["inlet_1", "inlet_2"]

    def test_create_inlet_list_inlet_list(self, mixer_frame):
        mixer_frame.fs.mix.config.inlet_list = ["foo", "bar"]

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()

        for i in inlet_list:
            assert i in ["foo", "bar"]

    def test_create_inlet_list_num_inlets(self, mixer_frame):
        mixer_frame.fs.mix.config.num_inlets = 3

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()

        for i in inlet_list:
            assert i in ["inlet_1", "inlet_2", "inlet_3"]

    def test_create_inlet_list_both_args_consistent(self, mixer_frame):
        mixer_frame.fs.mix.config.inlet_list = ["foo", "bar"]
        mixer_frame.fs.mix.config.num_inlets = 2

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()

        for i in inlet_list:
            assert i in ["foo", "bar"]

    def test_create_inlet_list_both_args_inconsistent(self, mixer_frame):
        mixer_frame.fs.mix.config.inlet_list = ["foo", "bar"]
        mixer_frame.fs.mix.config.num_inlets = 3

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        with pytest.raises(ConfigurationError):
            mixer_frame.fs.mix.create_inlet_list()

    def test_add_inlet_state_blocks(self, mixer_frame):
        mixer_frame.fs.mix.config.inlet_list = ["foo", "bar"]

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)

        assert isinstance(mixer_frame.fs.mix.foo_state, StateBlock)
        assert isinstance(mixer_frame.fs.mix.bar_state, StateBlock)

        assert len(inlet_blocks) == 2
        for i in inlet_blocks:
            assert isinstance(i, StateBlock)
            assert i.local_name in ["foo_state", "bar_state"]
            assert i[0].config.has_phase_equilibrium is False
            assert i[0].config.defined_state is True
            assert len(i[0].config) == 3

    def test_add_inlet_state_blocks_prop_pack_args(self, mixer_frame):
        mixer_frame.fs.mix.config.property_package_args = {"test": 1}
        mixer_frame.fs.mix.config.inlet_list = ["foo", "bar"]

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)

        assert isinstance(mixer_frame.fs.mix.foo_state, StateBlock)
        assert isinstance(mixer_frame.fs.mix.bar_state, StateBlock)

        assert len(inlet_blocks) == 2
        for i in inlet_blocks:
            assert isinstance(i, StateBlock)
            assert i.local_name in ["foo_state", "bar_state"]
            assert i[0].config.has_phase_equilibrium is False
            assert i[0].config.defined_state is True
            assert len(i[0].config) == 4
            assert i[0].config.test == 1

    def test_add_mixed_state_block(self, mixer_frame):
        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        assert isinstance(mixed_block, StateBlock)
        assert hasattr(mixer_frame.fs.mix, "mixed_state")
        assert not mixer_frame.fs.mix.mixed_state[0].config.has_phase_equilibrium
        assert mixer_frame.fs.mix.mixed_state[0].config.defined_state is False
        assert len(mixer_frame.fs.mix.mixed_state[0].config) == 3

    def test_add_mixed_state_block_prop_pack_args(self, mixer_frame):
        mixer_frame.fs.mix.config.property_package_args = {"test": 1}

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        assert isinstance(mixed_block, StateBlock)
        assert hasattr(mixer_frame.fs.mix, "mixed_state")
        assert not mixer_frame.fs.mix.mixed_state[0].config.has_phase_equilibrium
        assert mixer_frame.fs.mix.mixed_state[0].config.defined_state is False
        assert len(mixer_frame.fs.mix.mixed_state[0].config) == 4
        assert mixer_frame.fs.mix.mixed_state[0].config.test == 1

    def test_get_mixed_state_block(self, mixer_frame):
        mixer_frame.fs.sb = TestStateBlock(
                mixer_frame.fs.time,
                default={"parameters": mixer_frame.fs.pp})

        mixer_frame.fs.mix.config.mixed_state_block = mixer_frame.fs.sb

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        mixed_block = mixer_frame.fs.mix.get_mixed_state_block()

        assert mixed_block == mixer_frame.fs.sb

    def test_get_mixed_state_block_none(self, mixer_frame):
        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        with pytest.raises(BurntToast):
            mixer_frame.fs.mix.get_mixed_state_block()

    def test_get_mixed_state_block_mismatch(self, mixer_frame):
        mixer_frame.fs.sb = TestStateBlock(
                mixer_frame.fs.time,
                default={"parameters": mixer_frame.fs.pp})

        # Change parameters arg to create mismatch
        mixer_frame.fs.sb[0].config.parameters = None

        mixer_frame.fs.mix.config.mixed_state_block = mixer_frame.fs.sb

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        with pytest.raises(ConfigurationError):
            mixer_frame.fs.mix.get_mixed_state_block()

    # Test mixing equation methods
    def test_add_material_mixing_equations_pc(self, mixer_frame):
        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.componentPhase)

        assert isinstance(mixer_frame.fs.mix.material_mixing_equations,
                          Constraint)
        assert len(mixer_frame.fs.mix.material_mixing_equations) == 4

    def test_add_material_mixing_equations_pc_equilibrium(self, mixer_frame):
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.componentPhase)

        assert isinstance(mixer_frame.fs.mix.phase_equilibrium_generation,
                          Var)
        assert isinstance(mixer_frame.fs.mix.material_mixing_equations,
                          Constraint)
        assert len(mixer_frame.fs.mix.material_mixing_equations) == 4

    def test_add_material_mixing_equations_pc_equilibrium_not_supported(
            self, mixer_frame):
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        # Remove phase equilibrium list to trigger error
        mixer_frame.fs.pp.del_component(
                mixer_frame.fs.pp.phase_equilibrium_idx)

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        with pytest.raises(PropertyNotSupportedError):
            mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.componentPhase)

    def test_add_material_mixing_equations_tc(self, mixer_frame):
        mixer_frame.fs.mix.config.material_balance_type = \
            MaterialBalanceType.componentTotal

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.componentTotal)

        assert isinstance(mixer_frame.fs.mix.material_mixing_equations,
                          Constraint)
        assert len(mixer_frame.fs.mix.material_mixing_equations) == 2

    def test_add_material_mixing_equations_tc_equilibrium(self, mixer_frame):
        mixer_frame.fs.mix.config.material_balance_type = \
            MaterialBalanceType.componentTotal
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.componentTotal)

        assert isinstance(mixer_frame.fs.mix.material_mixing_equations,
                          Constraint)
        assert len(mixer_frame.fs.mix.material_mixing_equations) == 2

    def test_add_material_mixing_equations_t(self, mixer_frame):
        mixer_frame.fs.mix.config.material_balance_type = \
            MaterialBalanceType.total

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.total)

        assert isinstance(mixer_frame.fs.mix.material_mixing_equations,
                          Constraint)
        assert len(mixer_frame.fs.mix.material_mixing_equations) == 1

    def test_add_material_mixing_equations_t_equilibrium(self, mixer_frame):
        mixer_frame.fs.mix.config.material_balance_type = \
            MaterialBalanceType.total
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.total)

        assert isinstance(mixer_frame.fs.mix.material_mixing_equations,
                          Constraint)
        assert len(mixer_frame.fs.mix.material_mixing_equations) == 1

    def test_add_material_mixing_equations_e(self, mixer_frame):
        mixer_frame.fs.mix.config.material_balance_type = \
            MaterialBalanceType.elementTotal
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        with pytest.raises(ConfigurationError):
            mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.elementTotal)

    def test_add_material_mixing_equations_none(self, mixer_frame):
        mixer_frame.fs.mix.config.material_balance_type = \
            MaterialBalanceType.none
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_material_mixing_equations(
                inlet_blocks, mixed_block, MaterialBalanceType.none)

        assert not hasattr(mixer_frame.fs.mix, "material_mixing_equations")

    def test_add_energy_mixing_equations(self, mixer_frame):
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_energy_mixing_equations(inlet_blocks,
                                                       mixed_block)

        assert isinstance(mixer_frame.fs.mix.enthalpy_mixing_equations,
                          Constraint)
        assert len(mixer_frame.fs.mix.enthalpy_mixing_equations) == 1

    def test_add_pressure_minimization_equations(self, mixer_frame):
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_pressure_minimization_equations(inlet_blocks,
                                                               mixed_block)

        assert isinstance(mixer_frame.fs.mix.inlet_idx, Set)
        assert isinstance(mixer_frame.fs.mix.minimum_pressure, Var)
        assert len(mixer_frame.fs.mix.minimum_pressure) == 2
        assert isinstance(mixer_frame.fs.mix.eps_pressure, Param)
        assert isinstance(mixer_frame.fs.mix.minimum_pressure_constraint,
                          Constraint)
        assert len(mixer_frame.fs.mix.minimum_pressure) == 2
        assert isinstance(mixer_frame.fs.mix.mixture_pressure, Constraint)

    def test_add_pressure_equality_equations(self, mixer_frame):
        mixer_frame.fs.mix.config.has_phase_equilibrium = True

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_pressure_equality_equations(inlet_blocks,
                                                           mixed_block)

        assert isinstance(mixer_frame.fs.mix.pressure_equality_constraints,
                          Constraint)
        assert len(mixer_frame.fs.mix.pressure_equality_constraints) == 2

    def test_add_port_objects(self, mixer_frame):
        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_port_objects(inlet_list,
                                            inlet_blocks,
                                            mixed_block)

        assert isinstance(mixer_frame.fs.mix.inlet_1, Port)
        assert isinstance(mixer_frame.fs.mix.inlet_2, Port)
        assert isinstance(mixer_frame.fs.mix.outlet, Port)

    def test_add_port_objects_construct_ports_False(self, mixer_frame):
        mixer_frame.fs.mix.config.construct_ports = False

        mixer_frame.fs.mix._get_property_package()
        mixer_frame.fs.mix._get_indexing_sets()

        inlet_list = mixer_frame.fs.mix.create_inlet_list()
        inlet_blocks = mixer_frame.fs.mix.add_inlet_state_blocks(inlet_list)
        mixed_block = mixer_frame.fs.mix.add_mixed_state_block()

        mixer_frame.fs.mix.add_port_objects(inlet_list,
                                            inlet_blocks,
                                            mixed_block)

        assert hasattr(mixer_frame.fs.mix, "inlet_1") is False
        assert hasattr(mixer_frame.fs.mix, "inlet_2") is False
        assert hasattr(mixer_frame.fs.mix, "outlet") is False

    # -------------------------------------------------------------------------
    # Test build method
    @pytest.mark.build
    def test_build_default(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.pp = PhysicalParameterTestBlock()

        m.fs.mix = Mixer(default={"property_package": m.fs.pp})

        assert isinstance(m.fs.mix.material_mixing_equations, Constraint)
        assert len(m.fs.mix.material_mixing_equations) == 4
        assert hasattr(m.fs.mix, "phase_equilibrium_idx_ref") is False

        assert isinstance(m.fs.mix.enthalpy_mixing_equations, Constraint)
        assert len(m.fs.mix.enthalpy_mixing_equations) == 1

        assert isinstance(m.fs.mix.inlet_idx, Set)
        assert isinstance(m.fs.mix.minimum_pressure, Var)
        assert len(m.fs.mix.minimum_pressure) == 2
        assert isinstance(m.fs.mix.eps_pressure, Param)
        assert isinstance(m.fs.mix.minimum_pressure_constraint, Constraint)
        assert len(m.fs.mix.minimum_pressure) == 2
        assert isinstance(m.fs.mix.mixture_pressure, Constraint)

        assert isinstance(m.fs.mix.inlet_1, Port)
        assert isinstance(m.fs.mix.inlet_2, Port)
        assert isinstance(m.fs.mix.outlet, Port)

    @pytest.mark.build
    def test_build_phase_equilibrium(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.pp = PhysicalParameterTestBlock()

        m.fs.mix = Mixer(default={"property_package": m.fs.pp,
                                  "has_phase_equilibrium": True})

        assert isinstance(m.fs.mix.material_mixing_equations, Constraint)
        assert len(m.fs.mix.material_mixing_equations) == 4
        assert isinstance(m.fs.mix.phase_equilibrium_generation, Var)

        assert isinstance(m.fs.mix.enthalpy_mixing_equations, Constraint)
        assert len(m.fs.mix.enthalpy_mixing_equations) == 1

        assert isinstance(m.fs.mix.inlet_idx, Set)
        assert isinstance(m.fs.mix.minimum_pressure, Var)
        assert len(m.fs.mix.minimum_pressure) == 2
        assert isinstance(m.fs.mix.eps_pressure, Param)
        assert isinstance(m.fs.mix.minimum_pressure_constraint, Constraint)
        assert len(m.fs.mix.minimum_pressure) == 2
        assert isinstance(m.fs.mix.mixture_pressure, Constraint)

        assert isinstance(m.fs.mix.inlet_1, Port)
        assert isinstance(m.fs.mix.inlet_2, Port)
        assert isinstance(m.fs.mix.outlet, Port)

    @pytest.mark.build
    def test_build_phase_pressure_equality(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.pp = PhysicalParameterTestBlock()

        m.fs.mix = Mixer(default={
                "property_package": m.fs.pp,
                "momentum_mixing_type": MomentumMixingType.equality})

        assert isinstance(m.fs.mix.material_mixing_equations, Constraint)
        assert len(m.fs.mix.material_mixing_equations) == 4

        assert isinstance(m.fs.mix.enthalpy_mixing_equations, Constraint)
        assert len(m.fs.mix.enthalpy_mixing_equations) == 1

        assert isinstance(m.fs.mix.pressure_equality_constraints, Constraint)
        assert len(m.fs.mix.pressure_equality_constraints) == 2

        assert isinstance(m.fs.mix.inlet_1, Port)
        assert isinstance(m.fs.mix.inlet_2, Port)
        assert isinstance(m.fs.mix.outlet, Port)

    # -------------------------------------------------------------------------
    # Test models checks, initialize and release state methods
    def test_model_checks(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.pp = PhysicalParameterTestBlock()

        m.fs.mix = Mixer(default={
                "property_package": m.fs.pp,
                "momentum_mixing_type": MomentumMixingType.equality})

        m.fs.mix.model_check()

        assert m.fs.mix.inlet_1_state[0].check is True
        assert m.fs.mix.inlet_2_state[0].check is True
        assert m.fs.mix.mixed_state[0].check is True

    @pytest.mark.initialization
    def test_initialize(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.pp = PhysicalParameterTestBlock()
        m.fs.sb = TestStateBlock(m.fs.time, default={"parameters": m.fs.pp})

        m.fs.mix = Mixer(default={
                "property_package": m.fs.pp,
                "mixed_state_block": m.fs.sb})

        # Change one inlet pressure to check initialization calculations
        m.fs.mix.inlet_1_state[0].pressure = 8e4

        f = m.fs.mix.initialize(hold_state=True)

        assert m.fs.mix.inlet_1_state[0].init_test is True
        assert m.fs.mix.inlet_2_state[0].init_test is True
        assert m.fs.sb[0].init_test is True
        assert m.fs.mix.inlet_1_state[0].hold_state is True
        assert m.fs.mix.inlet_2_state[0].hold_state is True
        assert m.fs.sb[0].hold_state is False

        assert m.fs.sb[0].flow_mol_phase_comp["p1", "c1"].value == 4
        assert m.fs.sb[0].flow_mol_phase_comp["p1", "c2"].value == 4
        assert m.fs.sb[0].flow_mol_phase_comp["p2", "c1"].value == 4
        assert m.fs.sb[0].flow_mol_phase_comp["p2", "c2"].value == 4

        assert m.fs.sb[0].temperature.value == 300

        assert m.fs.sb[0].pressure.value == 8e4

        m.fs.mix.release_state(flags=f)

        assert m.fs.mix.inlet_1_state[0].hold_state is False
        assert m.fs.mix.inlet_2_state[0].hold_state is False
        assert m.fs.sb[0].hold_state is False

    @pytest.mark.ui
    def test_report(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.pp = PhysicalParameterTestBlock()
        m.fs.sb = TestStateBlock(m.fs.time, default={"parameters": m.fs.pp})

        m.fs.mix = Mixer(default={
                "property_package": m.fs.pp})

        m.fs.mix.report()


# -----------------------------------------------------------------------------
class TestBTX(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase": 'Liq'})

        m.fs.unit = Mixer(default={"property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "inlet_1")
        assert len(btx.fs.unit.inlet_1.vars) == 4
        assert hasattr(btx.fs.unit.inlet_1, "flow_mol")
        assert hasattr(btx.fs.unit.inlet_1, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet_1, "temperature")
        assert hasattr(btx.fs.unit.inlet_1, "pressure")

        assert hasattr(btx.fs.unit, "inlet_2")
        assert len(btx.fs.unit.inlet_2.vars) == 4
        assert hasattr(btx.fs.unit.inlet_2, "flow_mol")
        assert hasattr(btx.fs.unit.inlet_2, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet_2, "temperature")
        assert hasattr(btx.fs.unit.inlet_2, "pressure")

        assert hasattr(btx.fs.unit, "outlet")
        assert len(btx.fs.unit.outlet.vars) == 4
        assert hasattr(btx.fs.unit.outlet, "flow_mol")
        assert hasattr(btx.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.outlet, "temperature")
        assert hasattr(btx.fs.unit.outlet, "pressure")

        assert number_variables(btx) == 35
        assert number_total_constraints(btx) == 25
        assert number_unused_variables(btx) == 0

    def test_dof(self, btx):
        btx.fs.unit.inlet_1.flow_mol[0].fix(5)  # mol/s
        btx.fs.unit.inlet_1.temperature[0].fix(365)  # K
        btx.fs.unit.inlet_1.pressure[0].fix(2e5)  # Pa
        btx.fs.unit.inlet_1.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.inlet_1.mole_frac_comp[0, "toluene"].fix(0.5)

        btx.fs.unit.inlet_2.flow_mol[0].fix(1)  # mol/s
        btx.fs.unit.inlet_2.temperature[0].fix(300)  # K
        btx.fs.unit.inlet_2.pressure[0].fix(101325)  # Pa
        btx.fs.unit.inlet_2.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.inlet_2.mole_frac_comp[0, "toluene"].fix(0.5)

        assert degrees_of_freedom(btx) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, btx):
        orig_fixed_vars = fixed_variables_set(btx)
        orig_act_consts = activated_constraints_set(btx)

        btx.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(btx) == 0

        fin_fixed_vars = fixed_variables_set(btx)
        fin_act_consts = activated_constraints_set(btx)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, btx):
        assert (pytest.approx(6, abs=1e-3) ==
                value(btx.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(354.7, abs=1e-1) ==
                value(btx.fs.unit.outlet.temperature[0]))
        assert (pytest.approx(101325, abs=1e3) ==
                value(btx.fs.unit.outlet.pressure[0]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, btx):
        assert abs(value(btx.fs.unit.inlet_1.flow_mol[0] +
                         btx.fs.unit.inlet_2.flow_mol[0] -
                         btx.fs.unit.outlet.flow_mol[0])) <= 1e-6

        assert 1e-6 >= abs(value(
                btx.fs.unit.inlet_1.flow_mol[0] *
                btx.fs.unit.inlet_1_state[0].enth_mol_phase['Liq'] +
                btx.fs.unit.inlet_2.flow_mol[0] *
                btx.fs.unit.inlet_2_state[0].enth_mol_phase['Liq'] -
                btx.fs.unit.outlet.flow_mol[0] *
                btx.fs.unit.mixed_state[0].enth_mol_phase['Liq']))

    @pytest.mark.ui
    def test_report(self, btx):
        btx.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = Mixer(default={
                "property_package": m.fs.properties,
                "material_balance_type": MaterialBalanceType.componentTotal,
                "momentum_mixing_type": MomentumMixingType.equality})

        return m

    @pytest.mark.build
    def test_build(self, iapws):
        assert len(iapws.fs.unit.inlet_1.vars) == 3
        assert hasattr(iapws.fs.unit.inlet_1, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet_1, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet_1, "pressure")

        assert len(iapws.fs.unit.inlet_2.vars) == 3
        assert hasattr(iapws.fs.unit.inlet_2, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet_2, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet_2, "pressure")

        assert hasattr(iapws.fs.unit, "outlet")
        assert len(iapws.fs.unit.outlet.vars) == 3
        assert hasattr(iapws.fs.unit.outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.outlet, "pressure")

    def test_dof(self, iapws):
        iapws.fs.unit.inlet_1.flow_mol[0].fix(100)
        iapws.fs.unit.inlet_1.enth_mol[0].fix(4000)
        iapws.fs.unit.inlet_1.pressure[0].fix(101325)

        iapws.fs.unit.inlet_2.flow_mol[0].fix(100)
        iapws.fs.unit.inlet_2.enth_mol[0].fix(3500)

        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, iapws):
        orig_fixed_vars = fixed_variables_set(iapws)
        orig_act_consts = activated_constraints_set(iapws)

        iapws.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(iapws) == 0

        fin_fixed_vars = fixed_variables_set(iapws)
        fin_act_consts = activated_constraints_set(iapws)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, iapws):
        assert pytest.approx(200, abs=1e-5) == \
            value(iapws.fs.unit.outlet.flow_mol[0])

        assert pytest.approx(3750, abs=1e0) == \
            value(iapws.fs.unit.outlet.enth_mol[0])

        assert pytest.approx(101325, abs=1e2) == \
            value(iapws.fs.unit.outlet.pressure[0])

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, iapws):
        assert abs(value(iapws.fs.unit.inlet_1.flow_mol[0] +
                         iapws.fs.unit.inlet_2.flow_mol[0] -
                         iapws.fs.unit.outlet.flow_mol[0])) <= 1e-6

        assert abs(value(iapws.fs.unit.inlet_1.flow_mol[0] *
                         iapws.fs.unit.inlet_1.enth_mol[0] +
                         iapws.fs.unit.inlet_2.flow_mol[0] *
                         iapws.fs.unit.inlet_2.enth_mol[0] -
                         iapws.fs.unit.outlet.flow_mol[0] *
                         iapws.fs.unit.outlet.enth_mol[0])) <= 1e-6

    @pytest.mark.ui
    def test_report(self, iapws):
        iapws.fs.unit.report()


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = Mixer(default={"property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, sapon):
        assert len(sapon.fs.unit.inlet_1.vars) == 4
        assert hasattr(sapon.fs.unit.inlet_1, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet_1, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet_1, "temperature")
        assert hasattr(sapon.fs.unit.inlet_1, "pressure")

        assert len(sapon.fs.unit.inlet_2.vars) == 4
        assert hasattr(sapon.fs.unit.inlet_2, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet_2, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet_2, "temperature")
        assert hasattr(sapon.fs.unit.inlet_2, "pressure")

        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert number_variables(sapon) == 26
        assert number_total_constraints(sapon) == 10
        assert number_unused_variables(sapon) == 0

    def test_dof(self, sapon):
        sapon.fs.unit.inlet_1.flow_vol[0].fix(1e-3)
        sapon.fs.unit.inlet_1.temperature[0].fix(320)
        sapon.fs.unit.inlet_1.pressure[0].fix(101325)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.inlet_1.conc_mol_comp[0, "Ethanol"].fix(0.0)

        sapon.fs.unit.inlet_2.flow_vol[0].fix(1e-3)
        sapon.fs.unit.inlet_2.temperature[0].fix(300)
        sapon.fs.unit.inlet_2.pressure[0].fix(101325)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.inlet_2.conc_mol_comp[0, "Ethanol"].fix(0.0)

        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, sapon):
        orig_fixed_vars = fixed_variables_set(sapon)
        orig_act_consts = activated_constraints_set(sapon)

        sapon.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(sapon) == 0

        fin_fixed_vars = fixed_variables_set(sapon)
        fin_act_consts = activated_constraints_set(sapon)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, sapon):
        assert pytest.approx(2e-3, abs=1e-6) == \
            value(sapon.fs.unit.outlet.flow_vol[0])

        assert pytest.approx(55388.0, abs=1e0) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "H2O"])
        assert pytest.approx(100.0, abs=1e-3) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "NaOH"])
        assert pytest.approx(100.0, abs=1e-3) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"])
        assert pytest.approx(0.0, abs=1e-3) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"])
        assert pytest.approx(0.0, abs=1e-3) == value(
                sapon.fs.unit.outlet.conc_mol_comp[0, "Ethanol"])

        assert pytest.approx(310.0, abs=1e-1) == \
            value(sapon.fs.unit.outlet.temperature[0])

        assert pytest.approx(101325, abs=1e2) == \
            value(sapon.fs.unit.outlet.pressure[0])

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, sapon):
        assert abs(value(sapon.fs.unit.inlet_1.flow_vol[0] +
                         sapon.fs.unit.inlet_2.flow_vol[0] -
                         sapon.fs.unit.outlet.flow_vol[0])) <= 1e-6

        assert abs(value(
                sapon.fs.unit.inlet_1.flow_vol[0] *
                sapon.fs.properties.dens_mol*sapon.fs.properties.cp_mol *
                (sapon.fs.unit.inlet_1.temperature[0] -
                 sapon.fs.properties.temperature_ref) +
                sapon.fs.unit.inlet_2.flow_vol[0] *
                sapon.fs.properties.dens_mol*sapon.fs.properties.cp_mol *
                (sapon.fs.unit.inlet_2.temperature[0] -
                 sapon.fs.properties.temperature_ref) -
                sapon.fs.unit.outlet.flow_vol[0] *
                sapon.fs.properties.dens_mol*sapon.fs.properties.cp_mol *
                (sapon.fs.unit.outlet.temperature[0] -
                 sapon.fs.properties.temperature_ref))) <= 1e-3

    @pytest.mark.ui
    def test_report(self, sapon):
        sapon.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for equilibrium reactor unit model.
Authors: Andrew Lee
"""

import pytest
from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           value)
from idaes.core import (FlowsheetBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType)
from idaes.unit_models.equilibrium_reactor import EquilibriumReactor
from idaes.property_models.examples.saponification_thermo import (
                        SaponificationParameterBlock)
from idaes.property_models.examples.saponification_reactions import (
                        SaponificationReactionParameterBlock)
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock,
                                     ReactionParameterTestBlock)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.reactions = ReactionParameterTestBlock(default={
                            "property_package": m.fs.properties})

    m.fs.unit = EquilibriumReactor(default={
            "property_package": m.fs.properties,
            "reaction_package": m.fs.reactions})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 15

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == \
        MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == \
        EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.has_rate_reactions
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.has_equilibrium_reactions
    assert not m.fs.unit.config.has_phase_equilibrium
    assert not m.fs.unit.config.has_heat_of_reaction
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.reaction_package is m.fs.reactions


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(default={
                                "property_package": m.fs.properties})

        m.fs.unit = EquilibriumReactor(
                default={"property_package": m.fs.properties,
                         "reaction_package": m.fs.reactions,
                         "has_equilibrium_reactions": False,
                         "has_heat_transfer": True,
                         "has_heat_of_reaction": True,
                         "has_pressure_change": True})

        return m

    @pytest.mark.build
    def test_build(self, sapon):
        assert hasattr(sapon.fs.unit, "inlet")
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert hasattr(sapon.fs.unit, "outlet")
        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert hasattr(sapon.fs.unit, "rate_reaction_constraint")
        assert hasattr(sapon.fs.unit, "heat_duty")
        assert hasattr(sapon.fs.unit, "deltaP")

        assert number_variables(sapon) == 26
        assert number_total_constraints(sapon) == 16
        assert number_unused_variables(sapon) == 0

    def test_dof(self, sapon):
        sapon.fs.unit.inlet.flow_vol.fix(1.0e-03)
        sapon.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        sapon.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        sapon.fs.unit.inlet.temperature.fix(303.15)
        sapon.fs.unit.inlet.pressure.fix(101325.0)

        sapon.fs.unit.heat_duty.fix(0)
        sapon.fs.unit.deltaP.fix(0)

        assert degrees_of_freedom(sapon) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, sapon):
        orig_fixed_vars = fixed_variables_set(sapon)
        orig_act_consts = activated_constraints_set(sapon)

        sapon.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(sapon) == 0

        fin_fixed_vars = fixed_variables_set(sapon)
        fin_act_consts = activated_constraints_set(sapon)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, sapon):
        assert (pytest.approx(101325.0, abs=1e-2) ==
                value(sapon.fs.unit.outlet.pressure[0]))
        assert (pytest.approx(304.32, abs=1e-2) ==
                value(sapon.fs.unit.outlet.temperature[0]))
        assert (pytest.approx(0.01, abs=1e-2) ==
                value(sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, sapon):
        assert abs(value(sapon.fs.unit.inlet.flow_vol[0] -
                         sapon.fs.unit.outlet.flow_vol[0])) <= 1e-6
        assert (abs(value(sapon.fs.unit.inlet.flow_vol[0] *
                          sum(sapon.fs.unit.inlet.conc_mol_comp[0, j]
                              for j in sapon.fs.properties.component_list) -
                          sapon.fs.unit.outlet.flow_vol[0] *
                          sum(sapon.fs.unit.outlet.conc_mol_comp[0, j]
                              for j in sapon.fs.properties.component_list)))
                <= 1e-6)
        assert abs(value(sapon.fs.unit.inlet.flow_vol[0] *
                         sapon.fs.properties.cp_mol *
                         (sapon.fs.unit.inlet.temperature[0] -
                          sapon.fs.properties.temperature_ref) -
                         sapon.fs.unit.outlet.flow_vol[0] *
                         sapon.fs.properties.cp_mol *
                         (sapon.fs.unit.outlet.temperature[0] -
                          sapon.fs.properties.temperature_ref) +
                         sapon.fs.unit.heat_duty[0])) <= 1e-1

    @pytest.mark.ui
    def test_report(self, sapon):
        sapon.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for the cubic root finder external functions
"""

import pytest
import os

from numpy import logspace

from pyomo.environ import (ConcreteModel,
                           ExternalFunction,
                           value)
from pyomo.core.base.external import AMPLExternalFunction

from idaes import lib_directory

__author__ = "Andrew Lee"

# Set path to root finder .so file
_so = os.path.join(lib_directory, "cubic_roots.so")

# Set module level pyest marker
pytestmark = pytest.mark.cubic_root
prop_available = os.path.isfile(_so)

# Define parameters for different supported equations of state
EoS_param = {
        0: {'u': 2, 'w': -1},
        1: {'u': 1, 'w': 0}
        }

# Set paramter for number of points to test in reduced pressure and temperature
# Use Tr and Pr as A and B are linked
SAMPLES = 250
t_set = logspace(-1, 2, num=SAMPLES, base=10, endpoint=True)
p_set = logspace(-2, 2, num=SAMPLES, base=10, endpoint=True)

# Absolute tolerance for root finder checks
TOL = 1e-5
# Parameter indicating whether to test partial derivatives
TEST_DERS = True
# Relative finite difference step for partial derivatives
DEL = 1e-4
# Relative tolerance for accepting partial derivatives
FD_TOL = 5e-2


def between(y, x1, x2):
    return 0 > (y-x1)*(y-x2)


@pytest.fixture()
def root_finder():
    m = ConcreteModel()

    # Define external function methods
    m.proc_Z_liq = ExternalFunction(library=_so,
                                    function="ceos_z_liq")
    m.proc_Z_vap = ExternalFunction(library=_so,
                                    function="ceos_z_vap")
    m.proc_Z_liq_x = ExternalFunction(library=_so,
                                      function="ceos_z_liq_extend")
    m.proc_Z_vap_x = ExternalFunction(library=_so,
                                      function="ceos_z_vap_extend")

    return m


# TODO - Need tests for how external function behaves when given an invalid
# TODO - eos type. Currently seems to return a value, which probably should not
# TODO - happen.


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_roots_Z_liq(root_finder):
    for eos_type in [0, 1]:
        u = EoS_param[eos_type]["u"]
        w = EoS_param[eos_type]["w"]

        for T in t_set:
            for P in p_set:
                # Calculate A and B parameters from Tr and Pr
                A = 0.5*P/T**2
                B = 0.1*P/T

                # Get results of external function call
                f = root_finder.proc_Z_liq
                assert(isinstance(f, AMPLExternalFunction))
                Z, g, h = f.evaluate_fgh(args=(eos_type, A, B))

                # Calculate parameters of cubic
                c1 = 1
                c2 = -(1+B-u*B)
                c3 = (A-u*B-(u-w)*B**2)
                c4 = -(A*B+w*B**2+w*B**3)

                # Calculate residual and derivatives w.r.t. Z
                res = c1*Z**3 + c2*Z**2 + c3*Z + c4
                dz = 3*c1*Z**2 + 2*c2*Z + c3
                dz2 = 6*c1*Z + 2*c2

                try:
                    # Residual can be extemely sensitive to value of Z, so
                    # if residual fails tolerance, test size of Newton step to
                    # converge to root
                    assert (pytest.approx(0, abs=TOL) == res or
                            pytest.approx(0, abs=TOL) == value(res/dz))

                    # Check derivative signs to confirm correct root
                    assert dz >= 0  # Should always have non-negative slope

                    # Determine number of roots - calculate discriminant
                    dis = (18*c1*c2*c3*c4 - 4*c2**3*c4 + c2**2*c3**2 -
                           4*c1*c3**3 - 27*c1**2*c4**2)

                    if dis >= 0:
                        # Cubic has 2 or 3 real roots
                        # Second derivative should be non-positive
                        assert dz2 <= 0
                    # otherwise no need to check 2nd derivative

                    if TEST_DERS:
                        # Perform finite differences on A and B
                        ZAp, gAp, hAp = f.evaluate_fgh(
                                args=(eos_type, A*(1+DEL), B))
                        ZAm, gAm, hAm = f.evaluate_fgh(
                                args=(eos_type, A*(1-DEL), B))

                        ZBp, gBp, hBp = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1+DEL)))
                        ZBm, gBm, hBm = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1-DEL)))

                        # Check variance in Z values. A very large difference
                        # indicates a transition between single and multiple
                        # root regions, and hence that the partial derivatvies
                        # will be very sensitive.
                        # In these cases, skip the derivative tests.
                        if abs(ZAp - Z) > 1e-3 or abs(ZAm - Z) > 1e-3:
                            A_skip = True
                        else:
                            A_skip = False
                        if (abs(ZBp - Z) > 1e-3 or
                                abs(ZBm - Z) > 1e-3 or
                                abs(dis) < 1e-7):
                            B_skip = True
                        else:
                            B_skip = False

                        # Test gradient terms
                        # Calculate numerical first partial derivative
                        if not A_skip:
                            dZdA_p = (ZAp-Z)/(A*DEL)
                            dZdA_m = (Z-ZAm)/(A*DEL)

                        if not B_skip:
                            dZdB_p = (ZBp-Z)/(B*DEL)
                            dZdB_m = (Z-ZBm)/(B*DEL)

                        # Partial derivative w.r.t. EoS identifier
                        assert g[0] == 0

                        # Check that external function value lies within TOL of
                        # least one of the numerical values (delta+ or delta-),
                        # OR lies between the two numerical values and within
                        # 10*TOL of one of the numerical values
                        if not A_skip:
                            assert (
                                pytest.approx(dZdA_p, FD_TOL) == g[1] or
                                pytest.approx(dZdA_m, FD_TOL) == g[1] or
                                (between(g[1], dZdA_p, dZdA_m) and (
                                 pytest.approx(dZdA_p, 10*FD_TOL) == g[1] or
                                 pytest.approx(dZdA_m, 10*FD_TOL) == g[1])))

                        if not B_skip:
                            assert (
                                pytest.approx(dZdB_p, FD_TOL) == g[2] or
                                pytest.approx(dZdB_m, FD_TOL) == g[2] or
                                (between(g[2], dZdB_p, dZdB_m) and (
                                 pytest.approx(dZdB_p, 10*FD_TOL) == g[2] or
                                 pytest.approx(dZdB_m, 10*FD_TOL) == g[2])))

                        # Test hessian terms
                        # Calculate numerical second partial derivatives
                        if not A_skip:
                            d2ZdA2_p = (gAp[1]-g[1])/(A*DEL)
                            d2ZdA2_m = (g[1]-gAm[1])/(A*DEL)

                        if not B_skip:
                            d2ZdB2_p = (gBp[2]-g[2])/(B*DEL)
                            d2ZdB2_m = (g[2]-gBm[2])/(B*DEL)

                        if not A_skip and not B_skip:
                            d2ZdAB_p = (gBp[1]-g[1])/(B*DEL)
                            d2ZdAB_m = (g[1]-gBm[1])/(B*DEL)

                        # Partial derivatives w.r.t eos_type
                        assert h[0] == 0
                        assert h[1] == 0
                        assert h[3] == 0

                        if not A_skip:
                            assert (pytest.approx(d2ZdA2_p, FD_TOL) == h[2] or
                                    pytest.approx(d2ZdA2_m, FD_TOL) == h[2] or
                                    between(h[2], d2ZdA2_p, d2ZdA2_m))
                        if not A_skip and not B_skip:
                            assert (pytest.approx(d2ZdAB_p, FD_TOL) == h[4] or
                                    pytest.approx(d2ZdAB_m, FD_TOL) == h[4] or
                                    between(h[4], d2ZdAB_p, d2ZdAB_m))
                        if not B_skip:
                            # Second derivative w.r.t. B is very sensitive near
                            # point that roots disappear, and is at a
                            # maximum (or minimum) so skip tests if close to
                            # this point
                            assert (pytest.approx(d2ZdB2_p, FD_TOL) == h[5] or
                                    pytest.approx(d2ZdB2_m, FD_TOL) == h[5] or
                                    between(h[5], d2ZdB2_p, d2ZdB2_m))

                except AssertionError:
                    # Print values at failure and raise exception
                    print(eos_type, T, P, A, B, Z)
                    raise


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_roots_Z_vap(root_finder):
    for eos_type in [0, 1]:
        u = EoS_param[eos_type]["u"]
        w = EoS_param[eos_type]["w"]

        for T in t_set:
            for P in p_set:
                # Calculate A and B parameters from Tr and Pr
                A = 0.5*P/T**2
                B = 0.1*P/T

                # Get results of external function call
                f = root_finder.proc_Z_vap
                assert(isinstance(f, AMPLExternalFunction))
                Z, g, h = f.evaluate_fgh(args=(eos_type, A, B))

                # Calculate parameters of cubic
                c1 = 1
                c2 = -(1+B-u*B)
                c3 = (A-u*B-(u-w)*B**2)
                c4 = -(A*B+w*B**2+w*B**3)

                # Calculate residual and derivatives w.r.t. Z
                res = c1*Z**3 + c2*Z**2 + c3*Z + c4
                dz = 3*c1*Z**2 + 2*c2*Z + c3
                dz2 = 6*c1*Z + 2*c2

                try:
                    # Residual can be extemely sensitive to value of Z, so
                    # if residual fails tolerance, test size of Newton step to
                    # converge to root
                    assert (pytest.approx(0, abs=TOL) == res or
                            pytest.approx(0, abs=TOL) == value(res/dz))

                    # Check derivative signs to confirm correct root
                    assert dz >= 0  # Should always have non-negative slope

                    # Determine number of roots - calculate discriminant
                    dis = (18*c1*c2*c3*c4 - 4*c2**3*c4 + c2**2*c3**2 -
                           4*c1*c3**3 - 27*c1**2*c4**2)

                    if dis >= 0:
                        # Cubic has 2 or 3 real roots
                        # Second derivative should be non-negative
                        assert dz2 >= 0
                    # otherwise no need to check 2nd derivative

                    if TEST_DERS:
                        # Perform finite differences on A and B
                        ZAp, gAp, hAp = f.evaluate_fgh(
                                args=(eos_type, A*(1+DEL), B))
                        ZAm, gAm, hAm = f.evaluate_fgh(
                                args=(eos_type, A*(1-DEL), B))

                        ZBp, gBp, hBp = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1+DEL)))
                        ZBm, gBm, hBm = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1-DEL)))

                        # Check variance in Z values. A very large difference
                        # indicates a transition between single and multiple
                        # root regions, and hence that the partial derivatvies
                        # will be very sensitive.
                        # In these cases, skip the derivative tests.
                        if abs(ZAp - Z) > 1e-3 or abs(ZAm - Z) > 1e-3:
                            A_skip = True
                        else:
                            A_skip = False
                        if (abs(ZBp - Z) > 1e-3 or
                                abs(ZBm - Z) > 1e-3 or
                                abs(dis) < 1e-7):
                            B_skip = True
                        else:
                            B_skip = False

                        # Test gradient terms
                        # Calculate numerical first partial derivative
                        if not A_skip:
                            dZdA_p = (ZAp-Z)/(A*DEL)
                            dZdA_m = (Z-ZAm)/(A*DEL)

                        if not B_skip:
                            dZdB_p = (ZBp-Z)/(B*DEL)
                            dZdB_m = (Z-ZBm)/(B*DEL)

                        # Partial derivative w.r.t. EoS identifier
                        assert g[0] == 0

                        # Check that external function value lies within TOL of
                        # least one of the numerical values (delta+ or delta-),
                        # OR lies between the two numerical values and within
                        # 10*TOL of one of the numerical values
                        if not A_skip:
                            assert (
                                pytest.approx(dZdA_p, FD_TOL) == g[1] or
                                pytest.approx(dZdA_m, FD_TOL) == g[1] or
                                (between(g[1], dZdA_p, dZdA_m) and (
                                 pytest.approx(dZdA_p, 10*FD_TOL) == g[1] or
                                 pytest.approx(dZdA_m, 10*FD_TOL) == g[1])))

                        if not B_skip:
                            assert (
                                pytest.approx(dZdB_p, FD_TOL) == g[2] or
                                pytest.approx(dZdB_m, FD_TOL) == g[2] or
                                (between(g[2], dZdB_p, dZdB_m) and (
                                 pytest.approx(dZdB_p, 10*FD_TOL) == g[2] or
                                 pytest.approx(dZdB_m, 10*FD_TOL) == g[2])))

                        # Test hessian terms
                        # Calculate numerical second partial derivatives
                        if not A_skip:
                            d2ZdA2_p = (gAp[1]-g[1])/(A*DEL)
                            d2ZdA2_m = (g[1]-gAm[1])/(A*DEL)

                        if not B_skip:
                            d2ZdB2_p = (gBp[2]-g[2])/(B*DEL)
                            d2ZdB2_m = (g[2]-gBm[2])/(B*DEL)

                        if not A_skip and not B_skip:
                            d2ZdAB_p = (gBp[1]-g[1])/(B*DEL)
                            d2ZdAB_m = (g[1]-gBm[1])/(B*DEL)

                        # Partial derivatives w.r.t eos_type
                        assert h[0] == 0
                        assert h[1] == 0
                        assert h[3] == 0

                        if not A_skip:
                            assert (pytest.approx(d2ZdA2_p, FD_TOL) == h[2] or
                                    pytest.approx(d2ZdA2_m, FD_TOL) == h[2] or
                                    between(h[2], d2ZdA2_p, d2ZdA2_m))
                        if not A_skip and not B_skip:
                            assert (pytest.approx(d2ZdAB_p, FD_TOL) == h[4] or
                                    pytest.approx(d2ZdAB_m, FD_TOL) == h[4] or
                                    between(h[4], d2ZdAB_p, d2ZdAB_m))
                        if not B_skip:
                            # Second derivative w.r.t. B is very sensitive near
                            # point that roots disappear, and is at a
                            # maximum (or minimum) so skip tests if close to
                            # this point
                            assert (pytest.approx(d2ZdB2_p, FD_TOL) == h[5] or
                                    pytest.approx(d2ZdB2_m, FD_TOL) == h[5] or
                                    between(h[5], d2ZdB2_p, d2ZdB2_m))

                except AssertionError:
                    # Print values at failure and raise exception
                    print(eos_type, T, P, A, B, Z)
                    raise


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_roots_Z_liq_ext(root_finder):
    for eos_type in [0, 1]:
        u = EoS_param[eos_type]["u"]
        w = EoS_param[eos_type]["w"]

        for T in t_set:
            for P in p_set:
                # Calculate A and B parameters from Tr and Pr
                A = 0.5*P/T**2
                B = 0.1*P/T

                # Get results of external function call
                f = root_finder.proc_Z_liq_x
                assert(isinstance(f, AMPLExternalFunction))
                Z, g, h = f.evaluate_fgh(args=(eos_type, A, B))

                # Calculate parameters of cubic
                c1 = 1
                c2 = -(1+B-u*B)
                c3 = (A-u*B-(u-w)*B**2)
                c4 = -(A*B+w*B**2+w*B**3)

                det = c2**2 - 3*c3
                a = -(1.0/3.0)*(c2 + det**0.5)

                # Check to see if extension is triggered
                if det <= 0 or (a**3 + c2*a**2 + c3*a + c4) >= 0:
                    # Extension is not used
                    # Calculate residual and derivatives w.r.t. Z
                    res = c1*Z**3 + c2*Z**2 + c3*Z + c4
                    dz = 3*c1*Z**2 + 2*c2*Z + c3
                    dz2 = 6*c1*Z + 2*c2

                    try:
                        # Residual can be extemely sensitive to value of Z, so
                        # if residual fails tolerance, test size of Newton step
                        # to converge to root
                        assert (pytest.approx(0, abs=TOL) == res or
                                pytest.approx(0, abs=TOL) == value(res/dz))

                        # Check derivative signs to confirm correct root
                        assert dz >= 0  # Should always have non-negative slope

                        # Determine number of roots - calculate discriminant
                        dis = (18*c1*c2*c3*c4 - 4*c2**3*c4 + c2**2*c3**2 -
                               4*c1*c3**3 - 27*c1**2*c4**2)

                        if dis >= 0:
                            # Cubic has 2 or 3 real roots
                            # Second derivative should be non-positive
                            assert dz2 <= 0
                        # otherwise no need to check 2nd derivative

                    except AssertionError:
                        # Print values at failure and raise exception
                        print(eos_type, T, P, A, B, Z, det, a)
                        raise

                else:
                    # Extention is used, calculate extended root
                    c1x = 2*a
                    c2x = -c2 - 3.0*c1x
                    c3x = 3*c1x**2 + 2*c2*c1x + c3
                    c4x = c4 - 0.75*c1x**3 - 0.5*c2*c1x**2

                    # Calculate residual and derivatives w.r.t. Z_ext
                    res = c1*Z**3 + c2x*Z**2 + c3x*Z + c4x
                    dz = 3*c1*Z**2 + 2*c2x*Z + c3x

                    try:
                        # Residual can be extemely sensitive to value of Z, so
                        # if residual fails tolerance, test size of Newton step
                        # to converge to root
                        assert (pytest.approx(0, abs=TOL) == res or
                                pytest.approx(0, abs=TOL) == value(res/dz))

                        # Check derivative signs to confirm correct root
                        assert dz >= 0  # Should always have non-negative slope

                        # Determine number of roots - calculate discriminant
                        dis = (18*c1*c2x*c3x*c4x - 4*c2x**3*c4x +
                               c2x**2*c3x**2 - 4*c1*c3x**3 - 27*c1**2*c4x**2)

                        # Second derivative could be anything, don't check

                    except AssertionError:
                        # Print values at failure and raise exception
                        print(eos_type, T, P, A, B, Z)
                        raise

                if TEST_DERS:
                    try:
                        # Perform finite differences on A and B
                        ZAp, gAp, hAp = f.evaluate_fgh(
                                args=(eos_type, A*(1+DEL), B))
                        ZAm, gAm, hAm = f.evaluate_fgh(
                                args=(eos_type, A*(1-DEL), B))

                        ZBp, gBp, hBp = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1+DEL)))
                        ZBm, gBm, hBm = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1-DEL)))

                        # Check variance in Z values. A very large
                        # difference indicates a transition between
                        # single and multiple root regions, and hence that
                        # the partial derivatvies will be very sensitive.
                        # In these cases, skip the derivative tests.
                        if (abs(ZAp - Z) > 1e-3 or
                                abs(ZAm - Z) > 1e-3 or
                                abs(a) < 1e-1):
                            A_skip = True
                        else:
                            A_skip = False
                        if (abs(ZBp - Z) > 1e-3 or
                                abs(ZBm - Z) > 1e-3 or
                                abs(dis) < 1e-7 or
                                abs(a) < 1e-1):
                            B_skip = True
                        else:
                            B_skip = False

                        # Test gradient terms
                        # Calculate numerical first partial derivative
                        if not A_skip:
                            dZdA_p = (ZAp-Z)/(A*DEL)
                            dZdA_m = (Z-ZAm)/(A*DEL)

                        if not B_skip:
                            dZdB_p = (ZBp-Z)/(B*DEL)
                            dZdB_m = (Z-ZBm)/(B*DEL)

                        # Partial derivative w.r.t. EoS identifier
                        assert g[0] == 0

                        # Check that external function value lies within
                        # TOL of least one of the numerical values (delta+
                        # or delta-), OR lies between the two numerical
                        # values and within 10*TOL of one of the numerical
                        # values
                        if not A_skip:
                            assert (
                                pytest.approx(dZdA_p, FD_TOL) == g[1] or
                                pytest.approx(dZdA_m, FD_TOL) == g[1] or
                                (between(g[1], dZdA_p, dZdA_m) and (
                                 pytest.approx(dZdA_p, 10*FD_TOL) ==
                                 g[1] or
                                 pytest.approx(dZdA_m, 10*FD_TOL) ==
                                 g[1])))

                        if not B_skip:
                            assert (
                                pytest.approx(dZdB_p, FD_TOL) == g[2] or
                                pytest.approx(dZdB_m, FD_TOL) == g[2] or
                                (between(g[2], dZdB_p, dZdB_m) and (
                                 pytest.approx(dZdB_p, 10*FD_TOL) ==
                                 g[2] or
                                 pytest.approx(dZdB_m, 10*FD_TOL) ==
                                 g[2])))

                        # Test hessian terms
                        # Calculate numerical second partial derivatives
                        if not A_skip:
                            d2ZdA2_p = (gAp[1]-g[1])/(A*DEL)
                            d2ZdA2_m = (g[1]-gAm[1])/(A*DEL)

                        if not B_skip:
                            d2ZdB2_p = (gBp[2]-g[2])/(B*DEL)
                            d2ZdB2_m = (g[2]-gBm[2])/(B*DEL)

                        if not A_skip and not B_skip:
                            d2ZdAB_p = (gBp[1]-g[1])/(B*DEL)
                            d2ZdAB_m = (g[1]-gBm[1])/(B*DEL)

                        # Partial derivatives w.r.t eos_type
                        assert h[0] == 0
                        assert h[1] == 0
                        assert h[3] == 0

                        if not A_skip:
                            assert (
                                pytest.approx(d2ZdA2_p, FD_TOL) == h[2] or
                                pytest.approx(d2ZdA2_m, FD_TOL) == h[2] or
                                between(h[2], d2ZdA2_p, d2ZdA2_m))
                        if not A_skip and not B_skip:
                            assert (
                                pytest.approx(d2ZdAB_p, FD_TOL) == h[4] or
                                pytest.approx(d2ZdAB_m, FD_TOL) == h[4] or
                                between(h[4], d2ZdAB_p, d2ZdAB_m))
                        if not B_skip:
                            # Second derivative w.r.t. B is very sensitive
                            # near point that roots disappear, and is at a
                            # maximum (or minimum) so skip tests if close
                            # to this point
                            assert (
                                pytest.approx(d2ZdB2_p, FD_TOL) == h[5] or
                                pytest.approx(d2ZdB2_m, FD_TOL) == h[5] or
                                between(h[5], d2ZdB2_p, d2ZdB2_m))

                    except AssertionError:
                        # Print values at failure and raise exception
                        print(eos_type, T, P, A, B, Z)
                        raise


@pytest.mark.skipif(not prop_available,
                    reason="Cubic root finder not available")
def test_roots_Z_vap_ext(root_finder):
    for eos_type in [0, 1]:
        u = EoS_param[eos_type]["u"]
        w = EoS_param[eos_type]["w"]

        for T in t_set:
            for P in p_set:
                # Calculate A and B parameters from Tr and Pr
                A = 0.5*P/T**2
                B = 0.1*P/T

                # Get results of external function call
                f = root_finder.proc_Z_vap_x
                assert(isinstance(f, AMPLExternalFunction))
                Z, g, h = f.evaluate_fgh(args=(eos_type, A, B))

                # Calculate parameters of cubic
                c1 = 1
                c2 = -(1+B-u*B)
                c3 = (A-u*B-(u-w)*B**2)
                c4 = -(A*B+w*B**2+w*B**3)

                det = c2**2 - 3*c3
                a = -(1.0/3.0)*(c2 - det**0.5)

                # Check to see if extension is triggered
                if det <= 0 or (a**3 + c2*a**2 + c3*a + c4) <= 0:
                    # Extension is not used
                    # Calculate residual and derivatives w.r.t. Z
                    res = c1*Z**3 + c2*Z**2 + c3*Z + c4
                    dz = 3*c1*Z**2 + 2*c2*Z + c3
                    dz2 = 6*c1*Z + 2*c2

                    try:
                        # Residual can be extemely sensitive to value of Z, so
                        # if residual fails tolerance, test size of Newton step
                        # to converge to root
                        assert (pytest.approx(0, abs=TOL) == res or
                                pytest.approx(0, abs=TOL) == value(res/dz))

                        # Check derivative signs to confirm correct root
                        assert dz >= 0  # Should always have non-negative slope

                        # Determine number of roots - calculate discriminant
                        dis = (18*c1*c2*c3*c4 - 4*c2**3*c4 + c2**2*c3**2 -
                               4*c1*c3**3 - 27*c1**2*c4**2)

                        if dis >= 0:
                            # Cubic has 2 or 3 real roots
                            # Second derivative should be non-negative
                            assert dz2 >= 0
                        # otherwise no need to check 2nd derivative

                    except AssertionError:
                        # Print values at failure and raise exception
                        print(eos_type, T, P, A, B, Z)
                        raise

                else:
                    # Extention is used, calculate extended root
                    c1x = 2*a
                    c2x = -c2 - 3.0*c1x
                    c3x = 3*c1x**2 + 2*c2*c1x + c3
                    c4x = c4 - 0.75*c1x**3 - 0.5*c2*c1x**2

                    # Calculate residual and derivatives w.r.t. Z_ext
                    res = c1*Z**3 + c2x*Z**2 + c3x*Z + c4x
                    dz = 3*c1*Z**2 + 2*c2x*Z + c3x

                    try:
                        # Residual can be extemely sensitive to value of Z, so
                        # if residual fails tolerance, test size of Newton step
                        # to converge to root
                        assert (pytest.approx(0, abs=TOL) == res or
                                pytest.approx(0, abs=TOL) == value(res/dz))

                        # Check derivative signs to confirm correct root
                        assert dz >= 0  # Should always have non-negative slope

                        # Determine number of roots - calculate discriminant
                        dis = (18*c1*c2x*c3x*c4x - 4*c2x**3*c4x +
                               c2x**2*c3x**2 - 4*c1*c3x**3 - 27*c1**2*c4x**2)

                        # Second derivative could be anything, don't check

                    except AssertionError:
                        # Print values at failure and raise exception
                        print(eos_type, T, P, A, B, Z)
                        raise

                if TEST_DERS:
                    try:
                        # Perform finite differences on A and B
                        ZAp, gAp, hAp = f.evaluate_fgh(
                                args=(eos_type, A*(1+DEL), B))
                        ZAm, gAm, hAm = f.evaluate_fgh(
                                args=(eos_type, A*(1-DEL), B))

                        ZBp, gBp, hBp = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1+DEL)))
                        ZBm, gBm, hBm = f.evaluate_fgh(
                                    args=(eos_type, A, B*(1-DEL)))

                        # Check variance in Z values. A very large
                        # difference indicates a transition between
                        # single and multiple root regions, and hence that
                        # the partial derivatvies will be very sensitive.
                        # In these cases, skip the derivative tests.
                        if (abs(ZAp - Z) > 1e-3 or
                                abs(ZAm - Z) > 1e-3 or
                                abs(a) < 0.5):
                            A_skip = True
                        else:
                            A_skip = False
                        if (abs(ZBp - Z) > 1e-3 or
                                abs(ZBm - Z) > 1e-3 or
                                abs(dis) < 1e-7 or
                                abs(a) < 0.5):
                            B_skip = True
                        else:
                            B_skip = False

                        # Test gradient terms
                        # Calculate numerical first partial derivative
                        if not A_skip:
                            dZdA_p = (ZAp-Z)/(A*DEL)
                            dZdA_m = (Z-ZAm)/(A*DEL)

                        if not B_skip:
                            dZdB_p = (ZBp-Z)/(B*DEL)
                            dZdB_m = (Z-ZBm)/(B*DEL)

                        # Partial derivative w.r.t. EoS identifier
                        assert g[0] == 0

                        # Check that external function value lies within
                        # TOL of least one of the numerical values (delta+
                        # or delta-), OR lies between the two numerical
                        # values and within 10*TOL of one of the numerical
                        # values
                        if not A_skip:
                            assert (
                                pytest.approx(dZdA_p, FD_TOL) == g[1] or
                                pytest.approx(dZdA_m, FD_TOL) == g[1] or
                                (between(g[1], dZdA_p, dZdA_m) and (
                                 pytest.approx(dZdA_p, 10*FD_TOL) ==
                                 g[1] or
                                 pytest.approx(dZdA_m, 10*FD_TOL) ==
                                 g[1])))

                        if not B_skip:
                            assert (
                                pytest.approx(dZdB_p, FD_TOL) == g[2] or
                                pytest.approx(dZdB_m, FD_TOL) == g[2] or
                                (between(g[2], dZdB_p, dZdB_m) and (
                                 pytest.approx(dZdB_p, 10*FD_TOL) ==
                                 g[2] or
                                 pytest.approx(dZdB_m, 10*FD_TOL) ==
                                 g[2])))

                        # Test hessian terms
                        # Calculate numerical second partial derivatives
                        if not A_skip:
                            d2ZdA2_p = (gAp[1]-g[1])/(A*DEL)
                            d2ZdA2_m = (g[1]-gAm[1])/(A*DEL)

                        if not B_skip:
                            d2ZdB2_p = (gBp[2]-g[2])/(B*DEL)
                            d2ZdB2_m = (g[2]-gBm[2])/(B*DEL)

                        if not A_skip and not B_skip:
                            d2ZdAB_p = (gBp[1]-g[1])/(B*DEL)
                            d2ZdAB_m = (g[1]-gBm[1])/(B*DEL)

                        # Partial derivatives w.r.t eos_type
                        assert h[0] == 0
                        assert h[1] == 0
                        assert h[3] == 0

                        if not A_skip:
                            assert (
                                pytest.approx(d2ZdA2_p, FD_TOL) == h[2] or
                                pytest.approx(d2ZdA2_m, FD_TOL) == h[2] or
                                between(h[2], d2ZdA2_p, d2ZdA2_m))
                        if not A_skip and not B_skip:
                            assert (
                                pytest.approx(d2ZdAB_p, FD_TOL) == h[4] or
                                pytest.approx(d2ZdAB_m, FD_TOL) == h[4] or
                                between(h[4], d2ZdAB_p, d2ZdAB_m))
                        if not B_skip:
                            # Second derivative w.r.t. B is very sensitive
                            # near point that roots disappear, and is at a
                            # maximum (or minimum) so skip tests if close
                            # to this point
                            assert (
                                pytest.approx(d2ZdB2_p, FD_TOL) == h[5] or
                                pytest.approx(d2ZdB2_m, FD_TOL) == h[5] or
                                between(h[5], d2ZdB2_p, d2ZdB2_m))

                    except AssertionError:
                        # Print values at failure and raise exception
                        print(eos_type, T, P, A, B, Z)
                        raise



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Standard IDAES Equilibrium Reactor model.
"""

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import (is_physical_parameter_block,
                                    is_reaction_parameter_block)
from idaes.core.util.misc import add_object_reference

__author__ = "Andrew Lee"


@declare_process_block_class("EquilibriumReactor")
class EquilibriumReactorData(UnitModelBlockData):
    """
    Standard Equilibrium Reactor Unit Model Class
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
**default** = False. Equilibrium Reactors do not support dynamic behavior."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Equilibrium reactors do not have defined volume, thus
this must be False."""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("has_rate_reactions", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Rate reaction construction flag",
        doc="""Indicates whether terms for rate controlled reactions
should be constructed, along with constraints equating these to zero,
**default** - True.
**Valid values:** {
**True** - include rate reaction terms,
**False** - exclude rate reaction terms.}"""))
    CONFIG.declare("has_equilibrium_reactions", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Equilibrium reaction construction flag",
        doc="""Indicates whether terms for equilibrium controlled reactions
should be constructed,
**default** - True.
**Valid values:** {
**True** - include equilibrium reaction terms,
**False** - exclude equilibrium reaction terms.}"""))
    CONFIG.declare("has_phase_equilibrium", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Phase equilibrium term construction flag",
        doc="""Indicates whether terms for phase equilibrium should be
constructed, **default** - True.
**Valid values:** {
**True** - include phase equilibrium term,
**False** - exclude phase equlibirum terms.}"""))
    CONFIG.declare("has_heat_transfer", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Heat transfer term construction flag",
        doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}"""))
    CONFIG.declare("has_heat_of_reaction", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Heat of reaction term construction flag",
        doc="""Indicates whether terms for heat of reaction terms should be
constructed,
**default** - False.
**Valid values:** {
**True** - include heat of reaction terms,
**False** - exclude heat of reaction terms.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("reaction_package", ConfigValue(
        default=None,
        domain=is_reaction_parameter_block,
        description="Reaction package to use for control volume",
        doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}"""))
    CONFIG.declare("reaction_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing reaction packages",
        doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}"""))

    def build(self):
        """
        Begin building model.

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(EquilibriumReactorData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(default={
                "dynamic": self.config.dynamic,
                "has_holdup": self.config.has_holdup,
                "property_package": self.config.property_package,
                "property_package_args": self.config.property_package_args,
                "reaction_package": self.config.reaction_package,
                "reaction_package_args": self.config.reaction_package_args})

        # No need for control volume geometry

        self.control_volume.add_state_blocks(
                has_phase_equilibrium=self.config.has_phase_equilibrium)

        self.control_volume.add_reaction_blocks(
                has_equilibrium=self.config.has_equilibrium_reactions)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_rate_reactions=self.config.has_rate_reactions,
            has_equilibrium_reactions=self.config.has_equilibrium_reactions,
            has_phase_equilibrium=self.config.has_phase_equilibrium)

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_of_reaction=self.config.has_heat_of_reaction,
            has_heat_transfer=self.config.has_heat_transfer)

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        if self.config.has_rate_reactions:
            # Add equilibrium reactor performance equation
            @self.Constraint(self.flowsheet().config.time,
                             self.config.reaction_package.rate_reaction_idx,
                             doc="Rate reaction equilibrium constraint")
            def rate_reaction_constraint(b, t, r):
                # Set kinetic reaction rates to zero
                return b.control_volume.reactions[t].reaction_rate[r] == 0

        # Set references to balance terms at unit level
        if (self.config.has_heat_transfer is True and
                self.config.energy_balance_type != EnergyBalanceType.none):
            add_object_reference(self, "heat_duty", self.control_volume.heat)

        if (self.config.has_pressure_change is True and
                self.config.momentum_balance_type != 'none'):
            add_object_reference(self, "deltaP", self.control_volume.deltaP)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        return {"vars": var_dict}



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Resource representaitons.
"""
# stdlib
import abc
import argparse
from collections import namedtuple
from datetime import datetime
import getpass
import hashlib
import json
from json import JSONDecodeError
import logging
import os
import pathlib
import pprint
import re
import sys
from typing import List
import uuid

# third-party
import jsonschema
import pandas
import pendulum
import yaml

# local
from .util import datetime_timestamp

__author__ = "Dan Gunter"

_log = logging.getLogger(__name__)


class ProgLangExt:
    """Helper class to map from file extensions to
    names of the programming language.
    """

    _extmap = {
        "py": "Python",
        "pyc": "Python/compiled",
        "c": "C",
        "cpp": "C++",
        "cxx": "C++",
        "f": "FORTRAN",
        "f77": "FORTRAN",
        "f90": "FORTRAN",
        "js": "JavaScript",
        "jl": "Julia",
    }

    @classmethod
    def guess(cls, ext, default=None):
        ext = ext.lower()
        return cls._extmap.get(ext, default)


#: Constants for relation predicates
PR_DERIVED = "derived"  # derivedFrom
PR_CONTAINS = "contains"
PR_USES = "uses"
PR_VERSION = "version"
RELATION_PREDICATES = {PR_DERIVED, PR_CONTAINS, PR_USES, PR_VERSION}


TY_EXPERIMENT = "experiment"  #: Resource type for experiments
TY_TABULAR = "tabular_data"  #: Resource type for tabular data
TY_PROPERTY = "propertydb"  #: Resource type for property data
TY_FLOWSHEET = "flowsheet"  #: Resource type for a process flowsheet
TY_NOTEBOOK = "notebook"  #: Resource type for a Jupyter Notebook
TY_CODE = "code"  #: Resource type for source code
TY_SURRMOD = "surrogate_model"  #: Resource type for a surrogate model
TY_DATA = "data"  #: Resource type for generic data
TY_JSON = "json"  #: Resource type for JSON data
TY_OTHER = "other"  #: Resource type for unspecified type of resource
TY_RESOURCE_JSON = "resource_json"  #: Resource type for a JSON serialized resource

#: Constants for resource 'types'
RESOURCE_TYPES = {
    TY_EXPERIMENT,
    TY_TABULAR,
    TY_PROPERTY,
    TY_FLOWSHEET,
    TY_NOTEBOOK,
    TY_CODE,
    TY_SURRMOD,
    TY_DATA,
    TY_OTHER,
    TY_JSON,
    TY_RESOURCE_JSON,
}

# Constants for fields in stored relations
RR_PRED = "predicate"
RR_SUBJ = "subject"
RR_OBJ = "object"
RR_ID = "identifier"
RR_ROLE = "role"

RESOURCE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "id": "http://idaes.org",
    "definitions": {
        "SemanticVersion": {
            "title": "Version",
            "description": "Resource version using semantic versioning",
            "type": "array",
            "items": [
                {"type": "integer"},
                {"type": "integer"},
                {"type": "integer"},
                {"type": "string"},
            ],
            "minItems": 4,
        }
    },
    "type": "object",
    "properties": {
        "aliases": {"type": "array", "items": {"type": "string"}},
        "codes": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": [
                            "method",
                            "function",
                            "module",
                            "class",
                            "file",
                            "package",
                            "repository",
                            "notebook",
                        ],
                    },
                    "desc": {"type": "string"},
                    "name": {"type": "string"},
                    "language": {"type": "string"},
                    "idhash": {"type": "string"},
                    "location": {"type": "string"},
                    "version": {"$ref": "#/definitions/SemanticVersion"},
                },
                "required": ["name"],
            },
        },
        "collaborators": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "email": {"type": "string", "format": "email"},
                    "name": {"type": "string"},
                },
                "required": ["name"],
            },
        },
        "created": {"type": "number"},
        "creator": {
            "type": "object",
            "properties": {
                "email": {"type": "string", "format": "email"},
                "name": {"type": "string"},
            },
            "required": ["name"],
        },
        "data": {"type": "object"},
        "datafiles": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "desc": {"type": "string"},
                    "metadata": {"type": "object"},
                    "mimetype": {"type": "string"},
                    "path": {"type": "string"},
                    "sha1": {"type": "string"},
                    "is_copy": {"type": "boolean"},
                },
                "required": ["path"],
            },
        },
        "datafiles_dir": {"type": "string"},
        "desc": {"type": "string"},
        "id_": {"type": "string"},
        "modified": {"type": "number"},
        "relations": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    RR_PRED: {"type": "string", "enum": list(RELATION_PREDICATES)},
                    RR_ID: {"type": "string"},
                    RR_ROLE: {"type": "string", "enum": [RR_SUBJ, RR_OBJ]},
                },
                "required": [RR_PRED, RR_ID, RR_ROLE],
            },
        },
        "sources": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "date": {"type": "number"},
                    "doi": {"type": "string"},
                    "isbn": {"type": "string"},
                    "language": {"type": "string"},
                    "source": {"type": "string"},
                },
            },
        },
        "tags": {"type": "array", "items": {"type": "string"}},
        "type": {"type": "string", "enum": list(RESOURCE_TYPES)},
        "version_info": {
            "type": "object",
            "properties": {
                "created": {"type": "number"},
                "name": {"type": "string"},
                "version": {"$ref": "#/definitions/SemanticVersion"},
            },
        },
    },
    "required": ["id_"],
    "additionalProperties": False,
}


class Dict(dict):
    """Subclass of dict that has a 'dirty' bit.
    """

    def __init__(self, *args, **kwargs):
        super(Dict, self).__init__(*args, **kwargs)
        self._dirty = True

    def __setitem__(self, key, value):
        self._dirty = True
        super(Dict, self).__setitem__(key, value)

    def set_clean(self):
        self._dirty = False

    def is_dirty(self):
        return self._dirty


class Resource(object):
    """Core object for the Data Management Framework.
    """

    ID_FIELD = "id_"  #: Identifier field name constant
    ID_LENGTH = 32  #: Full-length of identifier
    TYPE_FIELD = "type"  #: Resource type field name constant

    def __init__(self, value: dict = None, type_: str = None):
        self._set_defaults()
        if value:
            _log.debug(f"update resource with values: {value}")
            self.v.update(value)
        if type_ is not None:
            self.v[self.TYPE_FIELD] = type_
        self._validator = jsonschema.Draft4Validator(RESOURCE_SCHEMA)
        self._validations = 0  # count validations; mostly for testing
        self.do_copy = self.is_tmp = False  # flags for copying datafiles

    def _set_defaults(self):
        now = date_float(pendulum.now())
        self.v = Dict(
            {
                self.ID_FIELD: identifier_str(),
                self.TYPE_FIELD: TY_OTHER,
                "aliases": [],
                "codes": [],
                "collaborators": [],
                "created": now,
                "modified": now,
                "creator": {"name": getpass.getuser()},
                "data": {},
                "datafiles": [],
                "datafiles_dir": "",
                "desc": "",
                "relations": [],
                "sources": [],
                "tags": [],
                "version_info": {"created": now, "version": (0, 0, 0), "name": ""},
            }
        )

    def _massage_values(self):
        try:
            # convert dates
            for item in self.v["sources"]:
                if not isinstance(item["date"], float):
                    item["date"] = date_float(item["date"])
            if not isinstance(self.v["created"], float):
                self.v["created"] = date_float(self.v["created"])
            if not isinstance(self.v["modified"], float):
                self.v["modified"] = date_float(self.v["modified"])
            if not isinstance(self.v["version_info"]["created"], float):
                self.v["version_info"]["created"] = date_float(
                    self.v["version_info"]["created"]
                )
            # convert versions
            if not isinstance(self.v["version_info"]["version"], list):
                self.v["version_info"]["version"] = version_list(
                    self.v["version_info"]["version"]
                )
            for i, code in enumerate(self.v["codes"]):
                if "version" in code:
                    if not isinstance(code["version"], list):
                        code["version"] = version_list(code["version"])
                        self.v["codes"][i] = code
        except (TypeError, ValueError, KeyError) as err:
            raise ValueError("While converting resource values: {}".format(err))
        self.v.set_clean()

    def validate(self):
        if self.v.is_dirty():
            self._massage_values()
            self._validator.validate(self.v)
            self._validations += 1

    # Some exceptions for communicating problems on import
    class InferResourceTypeError(Exception):
        pass

    class LoadResourceError(Exception):
        def __init__(self, inferred_type, msg):
            super().__init__(f"resource type '{inferred_type}': {msg}")

    @classmethod
    def from_file(
        cls, path: str, as_type: str = None, strict: bool = True, do_copy: bool = True
    ) -> "Resource":
        """Import resource from a file.

        Args:
            path: File path
            as_type: Resource type. If None/empty, then inferred from path.
            strict: If True, fail when file extension and contents don't match.
                    If False, always fall through to generic resource.
            do_copy: If True (the default), copy the files; else do not

        Raises:
            InferResourceTypeError: if resource type does not match inferred/specified
            LoadResourceError: if resource import failed
        """
        path = pathlib.Path(path)
        if as_type:
            if as_type == TY_RESOURCE_JSON:  # make sure resources validate
                try:
                    parsed = json.load(path.open())
                    jsonschema.Draft4Validator(RESOURCE_SCHEMA).validate(parsed)
                except (UnicodeDecodeError, JSONDecodeError):
                    raise ValueError("Resource is not well-formed JSON")
                except jsonschema.ValidationError as err:
                    raise ValueError(f"Resource does not match schema: {err}")
            else:
                parsed = None
        else:
            as_type, parsed = cls._infer_resource_type(path, strict)
        importer = cls._get_resource_importer(
            as_type, path, parsed=parsed, do_copy=do_copy
        )
        return importer.create()

    @classmethod
    def _infer_resource_type(cls, path: pathlib.Path, strict: bool):
        default_type = TY_OTHER
        try:
            if path.suffix == ".ipynb":
                return TY_NOTEBOOK, None
            if path.suffix == ".py":
                return TY_CODE, None
            if path.suffix == ".json":
                max_bytes = 1e6  # arbitrary limit
                # over max_bytes? generic
                file_size = path.stat().st_size
                if file_size > max_bytes:
                    _log.warning(
                        f"Not attempting to parse JSON, file size "
                        f"{file_size} > {max_bytes}"
                    )
                    return default_type, None
                # see if it's a Resource
                try:
                    parsed = json.load(path.open())
                except (UnicodeDecodeError, JSONDecodeError):
                    raise ValueError("File ending in '.json' is not valid JSON")
                try:
                    jsonschema.Draft4Validator(RESOURCE_SCHEMA).validate(parsed)
                except jsonschema.ValidationError:
                    return TY_JSON, parsed  # generic JSON
                return TY_RESOURCE_JSON, parsed
        except ValueError as err:
            if strict:
                raise cls.InferResourceTypeError(str(err))
            _log.warning(f"{err}: treating as generic file")
        return default_type, None

    @classmethod
    def _get_resource_importer(
        cls, type_: str, path: pathlib.Path, parsed=None, **kwargs
    ) -> "ResourceImporter":
        E = cls.LoadResourceError  # alias for exception raised
        if type_ == TY_NOTEBOOK:
            try:
                nb = json.load(open(str(path)))
            except (UnicodeDecodeError, JSONDecodeError):
                raise E(TY_NOTEBOOK, "not valid JSON")
            for key in "cells", "metadata", "nbformat":
                if key not in nb:
                    raise E(TY_NOTEBOOK, f"missing key: {key}")
            return JupyterNotebookImporter(path, **kwargs)
        if type_ == TY_CODE:
            language = ProgLangExt.guess(path.suffix, default="unknown")
            return CodeImporter(path, language, **kwargs)
        if type == TY_JSON:
            return JsonFileImporter(path, **kwargs)
        if type_ == TY_RESOURCE_JSON:
            return SerializedResourceImporter(path, parsed, **kwargs)
        return FileImporter(path, **kwargs)

    @property
    def id(self):
        """Get resource identifier.
        """
        return self.v[self.ID_FIELD]

    def set_id(self, value=None):
        self.v[self.ID_FIELD] = identifier_str(value)

    @property
    def name(self):
        """Get resource name (first alias).
        """
        try:
            nm = self.v["aliases"][0]
        except IndexError:
            nm = ""
        return nm

    @property
    def type(self):
        """Get resource type.
        """
        return self.v[self.TYPE_FIELD]

    @property
    def data(self):
        """Get JSON data for this resource.
        """
        return self.v["data"]

    @data.setter
    def data(self, value):
        """Set JSON data for this resource.
        """
        self.v["data"] = value

    def get_datafiles(self, mode="r"):
        """Generate readable file objects for 'datafiles' in resource.

        Args:
            mode (str): Mode for ``open()``
        Returns:
            generator: Generates ``file`` objects.
        """
        dfdir = self.v["datafiles_dir"]
        for datafile in self.v["datafiles"]:
            if not dfdir:
                path = datafile["path"]
            else:
                path = os.path.join(dfdir, datafile["path"])
            fp = open(path, mode=mode)
            yield fp

    def _repr_text_(self):
        return pprint.pformat(self.v, indent=2)

    def formatted_source(self) -> str:
        result = []
        for src in self.v['sources']:
            s = f"{src['source']}"
            if src['isbn']:
                s += f" ISBN: {src['isbn']}"
            if src['date']:
                s += f" Date: {src['date']}"
            result.append(s)
        return "\n".join(result)

#
# Function(s) to help creating [two-way] relations
# between resources
#


#: Provide attribute access to an RDF subject, predicate, object triple
Triple = namedtuple("Triple", "subject predicate object")


def create_relation(rel):
    """Create a relationship between two Resource instances.

    Relations are stored in both the `subject` and `object` resources, in
    the following way::

        If R = (subject)S, (predicate)P, and (object)O
        then store the following:
          In S.relations: {predicate: P, identifier:O.id, role:subject}
          In O.relations: {predicate: P, identifier:S.id, role:object}

    Args:
        rel (Triple): Relation triple. The 'subject' and 'object' parts
                    should be :class:`Resource`, and the 'predicate' should
                    be a simple string.
    Returns:
        None
    Raises:
        ValueError: if this relation already exists in the subject or
                    object resource, or the predicate is not in the list
                    of valid ones in RELATION_PREDICATES
    """
    if rel.predicate not in RELATION_PREDICATES:
        raise ValueError(
            'Bad predicate: "{}" not in: {}'.format(
                rel.predicate, ", ".join(list(RELATION_PREDICATES))
            )
        )
    rel_d = {
        RR_PRED: rel.predicate,
        RR_ID: rel.object.v[Resource.ID_FIELD],
        RR_ROLE: RR_SUBJ,
    }
    if rel_d in rel.subject.v["relations"]:
        raise ValueError("Duplicate relation for subject: {}".format(rel))
    rel.subject.v["relations"].append(rel_d)
    rel_d = {
        RR_PRED: rel.predicate,
        RR_ID: rel.subject.v[Resource.ID_FIELD],
        RR_ROLE: RR_OBJ,
    }
    # note: hard for this to happen unless the relation was added manually
    if rel_d in rel.object.v["relations"]:
        raise ValueError("Duplicate relation for object: {}".format(rel))
    rel.object.v["relations"].append(rel_d)


def create_relation_args(*args):
    """Syntactic sugar to take 3 args instead of a Triple.
    """
    return create_relation(Triple(*args))


def triple_from_resource_relations(id_, rrel):
    """Create a Triple from one entry in resource['relations'].

    Args:
        id_ (str): Identifier of the containing resource.
        rrel (dict): Stored relation with three keys, see `create_relation()`.
    Return:
        Triple: A triple
    """
    if rrel[RR_ROLE] == RR_SUBJ:
        rel = Triple(id_, rrel[RR_PRED], rrel[RR_ID])
    else:
        rel = Triple(rrel[RR_ID], rrel[RR_PRED], id_)
    return rel


#
# Some handy-dandy conversion functions.
#


def date_float(value):
    """Convert a date to a floating point seconds since the UNIX epoch.
    """

    def bad_date(e):
        raise ValueError('Cannot convert date "{}" to float: {}'.format(value, e))

    dt, usec = None, 0
    if isinstance(value, pendulum.Pendulum):
        return value.timestamp()
    elif isinstance(value, datetime):
        dt = value
    elif isinstance(value, tuple):
        try:
            dt = datetime(*value)
        except TypeError as err:
            bad_date(err)
    elif isinstance(value, str):
        try:
            dt = pendulum.parse(value)
        except pendulum.exceptions.ParserError as err:
            bad_date(err)
    elif isinstance(value, float) or isinstance(value, int):
        try:
            dt = datetime.fromtimestamp(value)
        except ValueError as err:
            bad_date(err)
    if dt is None:
        raise ValueError('Cannot convert date, value is "None"')
    return datetime_timestamp(dt) + usec  # just a float


# def isoformat(ts):
#     return datetime.fromtimestamp(ts).isoformat()


def version_list(value):
    """Semantic version.

    Three numeric identifiers, separated by a dot.
    Trailing non-numeric characters allowed.

    Inputs, string or tuple, may have less than three numeric
    identifiers, but internally the value will be padded with
    zeros to always be of length four.

    A leading dash or underscore in the trailing non-numeric characters
    is removed.

    Some examples of valid inputs and how they translate to 4-part versions:

    .. testsetup:: version_list

        from idaes.dmf.resource import version_list

    .. doctest:: version_list

        >>> version_list('1')
        [1, 0, 0, '']
        >>> version_list('1.1')
        [1, 1, 0, '']
        >>> version_list('1a')
        [1, 0, 0, 'a']
        >>> version_list('1.12.1')
        [1, 12, 1, '']
        >>> version_list('1.12.13-1')
        [1, 12, 13, '1']

    Some examples of invalid inputs:

    .. doctest:: version_list

        >>> for bad_input in ('rc3',      # too short
        ...                   '1.a.1.',   # non-number in middle
        ...                   '1.12.13.x' # too long
        ...     ):
        ...     try:
        ...         version_list(bad_input)
        ...     except ValueError:
        ...         print(f"failed: {bad_input}")
        ...
        failed: rc3
        failed: 1.a.1.
        failed: 1.12.13.x


    Returns:
        list: [major:int, minor:int, debug:int, release-type:str]
    """

    def bad_version(v):
        raise ValueError("Bad version: {}".format(v))

    ver = ()
    if isinstance(value, list) or isinstance(value, tuple):
        ver = value
    elif isinstance(value, str):
        ver = value.split(".", 2)
    elif isinstance(value, int):
        ver = (value, 0, 0)
    else:
        bad_version(value)
    if len(ver) < 1:
        bad_version(value)
    verlist = []
    # leading version numbers
    for i in range(len(ver) - 1):
        try:
            verlist.append(int(ver[i]))
        except ValueError:
            bad_version(value)
    # last version number
    s = ver[-1]
    extra = ""
    if isinstance(s, int):
        verlist.append(s if len(verlist) < 3 else str(s))
    elif isinstance(s, str):
        if s:
            m = re.match("([0-9]+)?(.*)", s)
            if m.group(1) is not None:
                verlist.append(int(m.group(1)))
            extra = m.group(2)
    else:  # last version must be int or str
        bad_version(value)
    # must have at least one numbered version
    if len(verlist) == 0:
        bad_version(value)
    # pad with zeros, and add non-numeric ID
    while len(verlist) < 3:
        verlist.append(0)
    if extra and extra[0] == ".":
        # cannot start extra version with '.'
        bad_version(value)
    if extra and extra[0] in ("-", "_"):
        extra = extra[1:]
    verlist.append(extra)
    return verlist


def format_version(values):
    s = "{}.{}.{}".format(*values[:3])
    if len(values) > 3 and values[3]:
        s += "-{}".format(values[3])
    return s


def identifier_str(value=None, allow_prefix=False):
    """Generate or validate a unique identifier.

    If generating, you will get a UUID in hex format

    .. testsetup:: idstr

        from idaes.dmf.resource import identifier_str

    .. doctest:: idstr

        >>> identifier_str()  #doctest: +ELLIPSIS
        '...'

    If validating, anything that is not 32 lowercase letters
    or digits will fail.

    .. doctest:: idstr

        >>> identifier_str('A' * 32)   #doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ValueError: Bad format for identifier "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA":
        must match regular expression "[0-9a-f]{32}"

    Args:
        value (str): If given, validate that it is a 32-byte str
                     If not given or None, set new random value.
    Raises:
        ValuError, if a value is given, and it is invalid.
    """
    # regular expression for identifier: hex string len=32
    if allow_prefix:
        id_expr = "[0-9a-f]{1,32}"
    else:
        id_expr = "[0-9a-f]{32}"
    if value is None:
        value = uuid.uuid4().hex
    elif not re.match(id_expr, value):
        raise ValueError(
            'bad format for identifier "{}": must match '
            'regular expression "{}"'.format(value, id_expr)
        )
    return value


class TidyUnitData:
    """Handle "tidy data" with per-column units.

    This can be used to convert from a simple dictionary/json
    representation like this::

            {
              "variables": ["compound", "pressure"],
              "units": [null|None, "Pa"],
              "observations": [
                ["benzene", 4890000.0],
                ...etc..
              ]
            }

    into a pandas DataFrame. A convenience method is provided for returning
    the data in a format easily dealt with when creating unit block parameters.
    Note that the keys in the preceding dictionary match the names of the
    parameters in the constructor (so you can pass this directly in as '**arg').

    Attributes:
        units (list): Units for each column, None where no units are defined
        table (pandas.DataFrame): The observation data
    """
    def __init__(self, data: dict = None, variables: List = None, units: List = None,
                 observations: List = None):
        """Constructor.

        Args:
            data: Optional, data dict (overrides variables/units/observations)
            variables: List of variables (table header)
            units: List of units, or None, same length as `variables`
            observations: Rows of the body of the table
        Raises:
            ValueError: For bad `data` (missing keys, not dict, etc.), or mismatches
                        in lengths of various pieces.
        """
        if data:
            try:
                variables, units, observations = (data['variables'], data['units'],
                                                  data['observations'])
            except KeyError as err:
                raise ValueError(f"Missing expected key in `data` param: {err}")
            except TypeError as err:
                raise ValueError(f"Bad value for `data` param: {err}")
        n = len(variables)
        if n == 0:
            self.df, self.units = pandas.DataFrame(), ()
            return
        if len(units) != n:
            raise ValueError(f"Length of units {len(units)} "
                             f"must match length of header ({n})")
        self.units = units
        self.table = pandas.DataFrame(data=observations, columns=variables)

    @property
    def param_data(self) -> dict:
        """Data in a form easily consumed by unit block params.

        The dictionary returned is like ``{ (key1, key2, ..): value }``,
        where the keys are values from all columns except the last,
        and the value is the last column.
        """
        d = {}
        for row in self.table.itertuples():
            key = tuple(row[1:-1])
            if len(key) == 1:
                key = key[0]
            value = row[-1]
            d[key] = value
        return d

#
# Import Resource of varying types from file
#


class ResourceImporter(abc.ABC):
    """Base class for Resource importers.
    """

    def __init__(self, path: pathlib.Path, do_copy: bool = None):
        self._path = path
        self._do_copy = do_copy

    def create(self) -> Resource:
        """Factory method.
        """
        r = self._create()
        r.validate()
        return r

    @abc.abstractmethod
    def _create(self) -> Resource:
        pass

    def _add_datafiles(self, r):
        abspath = str(self._path.absolute())
        file_hash = self._hash_file(abspath)
        r.v["datafiles"].append(
            {
                "desc": self._path.name,
                "path": abspath,
                "do_copy": self._do_copy,
                "sha1": file_hash,
            }
        )

    def _hash_file(self, path):
        blksz, h = 1 << 16, hashlib.sha1()
        with open(path, "rb") as f:
            blk = f.read(blksz)
            while blk:
                h.update(blk)
                blk = f.read(blksz)
        return h.hexdigest()


class JupyterNotebookImporter(ResourceImporter):
    def _create(self) -> Resource:
        r = Resource(type_=TY_NOTEBOOK)
        # XXX: add notebook 'metadata' as FilePath metadata attr
        self._add_datafiles(r)
        r.v["desc"] = self._path.name
        return r


class CodeImporter(ResourceImporter):
    def __init__(self, path, language, **kwargs):
        super().__init__(path, **kwargs)
        self.language = language

    def _create(self) -> Resource:
        r = Resource(type_=TY_CODE)
        r.v["codes"].append(
            {"name": self._path.name, "language": self.language, "type": "module"}
        )
        self._add_datafiles(r)
        r.v["desc"] = self._path.name
        return r


class FileImporter(ResourceImporter):
    def _create(self) -> Resource:
        r = Resource(type_=TY_DATA)
        self._add_datafiles(r)
        r.v["desc"] = str(self._path)
        return r


class JsonFileImporter(ResourceImporter):
    def _create(self) -> Resource:
        r = Resource(type_=TY_JSON)
        self._add_datafiles(r)
        r.v["desc"] = str(self._path)
        return r


class SerializedResourceImporter(ResourceImporter):
    def __init__(self, path, parsed, **kwargs):
        super().__init__(path, **kwargs)
        self.parsed = parsed

    def _create(self) -> Resource:
        r = Resource(value=self.parsed)
        return r


#
# Fill any JSON-schema-constrained instance with
# dummy values
#


def add_dummy_values(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    dummy_for_type = {
        "object": {},
        "boolean": False,
        "null": None,
        "number": 0,
        "integer": 0,
        "string": "",
        "any": "",
    }

    def set_defaults(validator, properties, instance, schema):
        for prop, subschema in properties.items():
            # print(f"@@ at prop={prop}")
            if "$ref" in subschema:
                refkey = subschema["$ref"].split("/")[-1]
                subschema = RESOURCE_SCHEMA["definitions"][refkey]
                # print(f"@@ ref {refkey} resolved, schema={subschema}")
            if "enum" in subschema:
                value = subschema["enum"][0]
            elif subschema["type"] == "array":
                # print("@@ array subschema")
                if "minItems" in subschema:
                    value = [
                        dummy_for_type[item["type"]] for item in subschema["items"]
                    ]
                elif subschema["items"]["type"] == "object":
                    value = [{}]
                else:
                    value = []
                # print(f"@@ array value = {value}")
            else:
                value = dummy_for_type[subschema["type"]]
            instance.setdefault(prop, value)
        for error in validate_properties(validator, properties, instance, schema):
            yield error

    return jsonschema.validators.extend(validator_class, {"properties": set_defaults})


# Create dummy resource

DummyValueValidator = add_dummy_values(jsonschema.Draft4Validator)
DUMMY_RESOURCE = {}
DummyValueValidator(RESOURCE_SCHEMA).validate(DUMMY_RESOURCE)
# print("@@ dummy resource:\n{}".format(DUMMY_RESOURCE))


def schema_as_yaml():
    """Export resource schema as YAML suitable for embedding into, e.g.,
       an OpenAPI spec.
    """
    return yaml.dump(RESOURCE_SCHEMA)


#
# Things to do if run as a script
#


if __name__ == "__main__":
    # parse command line
    ap = argparse.ArgumentParser()
    actions = {
        "json_schema": "print resource schema as JSON",
        "yaml_schema": "print resource schema as YAML",
    }
    ap.add_argument(
        "action", help="Action when run as a script", choices=list(actions.keys())
    )
    args = ap.parse_args()
    # take appropriate action
    if args.action == "json_schema":
        json.dump(RESOURCE_SCHEMA, sys.stdout, indent=2)
    elif args.action == "yaml_schema":
        print(schema_as_yaml())
    else:
        print("nothing to do")
    # exit
    sys.exit(0)



"""
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
"""
__all__ = ["ripemodel", "ems", "rspace", "sharedata", "debug", 
           "powerlawp5",
           "powerlaw2",
           "powerlaw3",
           "powerlaw4",
           "avrami2",
           "avrami3",
           "avrami4",
           "avrami5",
           "randomnuc",
           "ptompkins",
           "jander",
           "antijander",
           "valensi",
           "parabolic",
           "gb3d",
           "zlt",
           "grain",
           "massact",
           "massactm",
           "getmechs"]

from .main import ripemodel, ripewrite, print_results     # noqa: F401
from .shared import rspace, sharedata, debug    # noqa: F401
from .atermconstruct import makeaterm, formatinputs, checkargs, normalizefeatures     # noqa: F401
from .kinforms import lin, linjac, arr, arrjac, refarr, refarrjac   # noqa: F401
from .mechs import powerlawp5, powerlaw2, powerlaw3, powerlaw4, \
                   avrami2, avrami3, avrami4, avrami5, randomnuc, \
                   ptompkins, jander, antijander, valensi, \
                   parabolic, gb3d, zlt, grain, \
                   getmechs, massactm                            # noqa: F401
from .genpyomo import ripeomo                         # noqa: F401
from .targets import doalamo, dopwalamo, gentargets, sstargets, dynamictargets   # noqa: F401
from .confinv import confinv                          # noqa: F401
from .emsampling import constructmodel, ems                       # noqa: F401
from .checkoptions import checkoptions                     # noqa: F401    
from .bounds import stoich_cons, count_neg, get_bounds                           # noqa: F401



#!/usr/bin/env python
"""
Institute for the Design of Advanced Energy Systems
"""
from pathlib import Path
import os
import sys
from setuptools import setup, find_packages


def warn(s):
    sys.stderr.write("*** WARNING *** {}\n".format(s))


def get_version():
    code_file = os.path.join("idaes", "ver.py")
    code = open(code_file).read()
    local_namespace = {}
    exec(code, {}, local_namespace)
    return local_namespace["__version__"]


NAME = "idaes-pse"
VERSION = get_version()
README = open("README.md").read()
README = README[README.find("#") :]  # ignore everything before title


def rglob(path, glob):
    """Return list of paths from `path` matching `glob`.
    """
    p = Path(path)
    return list(map(str, p.rglob(glob)))


kwargs = dict(
    zip_safe=False,
    name=NAME,
    version=VERSION,
    packages=find_packages(),
    # Put abstract (non-versioned) deps here.
    # Concrete dependencies go in requirements[-dev].txt
    install_requires=[
        # idaes core / dmf
        "backports.shutil_get_terminal_size",
        "bokeh==0.12.9",
        "bunch",
        "click",
        "colorama",
        "humanize",
        "jupyter",
        "lxml",
        "matplotlib",
        "mock",
        "nbconvert",
        "nbformat",
        "numpy",
        "networkx",
        "pandas",
        "pendulum==1.4.4",
        "pint",
        "psutil",
        "pyomo",
        "pytest",
        "pyutilib",
        "pyyaml",
        "sympy",
        "tinydb",
        "toml",
        # alamopy
        # <nothing>
        # ripe
        # <nothing>
        # helmet
        "rbfopt",
    ],
    entry_points={
        "console_scripts": [
            "dmf = idaes.dmf.cli:base_command",
            "idaes = idaes.commands.base:command_base",
        ]
    },
    extras_require={
        # For developers. Only installed if [dev] is added to package name
        "dev": [
            "alabaster>=0.7.7",
            "coverage",
            "flake8",
            "flask>=1.0",
            "flask-bower",
            "flask-restful",
            "jsonschema",
            "jupyter_contrib_nbextensions",
            "mock",
            "pylint",
            "pytest-cov",
            "python-coveralls",
            "snowballstemmer==1.2.1",
            "sphinx-rtd-theme>=0.1.9",
            "sphinxcontrib-napoleon>=0.5.0",
            "sphinx-argparse",
        ]
    },
    package_data={
        # If any package contains these files, include them:
        "": [
            "*.template",
            "*.json",
            "*.yaml",
            "*.svg",
            "*.png",
            "*.jpg",
            "*.csv",
            "*.ipynb",
        ]
    },
    include_package_data=True,
    data_files=[],
    maintainer="Keith Beattie",
    maintainer_email="ksbeattie@lbl.gov",
    url="https://idaes.org",
    license="BSD ",
    platforms=["any"],
    description="IDAES Process Systems Engineering Framework",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=[NAME, "energy systems", "chemical engineering", "process modeling"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)

setup(**kwargs)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

"""
This script downloads a python file from pyomo.org that will allow us to update 
the workshop material easily during a workshop.

The file install_idaes_workshop_materials.py is downloaded from pyomo.org, imported,
 and the method execute() is then called to do whatever actions are necessary.

Note:
- update_workshop_materials.py is a module in the idaes/util folder that does the work
- update_workshop_materials.ipynb is a Jupyter notebook in the examples/workshops folder
  that calls this module
- install_idaes_workshop_materials.py is posted on a site (for now pyomo.org, but later 
  could be a repository on github).

1) JupyterHub: User executes update_workshop_materials.ipynb which calls to update_workshop_materials.py in idaes/util
2) update_workshop_materials.py downloads another python file (install_idaes_workshop_materials.py)
   from pyomo.org and calls "execute()" from that module.
3) install_idaes_workshop_materials.py does whatever is necessary to get the workshop materials 
   into the user folder (and perform any updates necessary).

TODO: This should probably be changed to get a zip file from an IDAES repository rather than install_idaes_workshop_materials.py from pyomo.org
"""

import os
import logging
import pyomo.common.fileutils as futils
import pyomo.common.download as dload
from pyutilib.misc import import_file

_install_idaes_workshop_materials_url = 'http://www.pyomo.org/s/install_idaes_workshop_materials.py'

def download_install_module():
    """
    Downloads install_idaes_workshop_materials.py from pyomo.org
    """
    download_dir = futils.this_file_dir()
    download_dir = os.path.join(download_dir, '../../examples/workshops')
    download_dest = os.path.join(download_dir, 'install_idaes_workshop_materials.py')
    
    print("\n\n"
          "#######################################################################################\n"
          "# Downloading: {}\n"
          "# to: {}\n"
          "#######################################################################################\n"
          "".format(_install_idaes_workshop_materials_url, download_dest)
    )

    if not os.path.isdir(download_dir):
        raise NameError('Unable to locate download directory: {}'.format(download_dir))

    try:
        downloader = dload.FileDownloader()
        downloader.set_destination_filename(download_dest)
        downloader.get_binary_file(_install_idaes_workshop_materials_url)
    except:
        print("\n\n***\nFailed to download: {}\n***".format(_install_idaes_workshop_materials_url))
        raise

    print('... download complete')
    return download_dest

def import_install_module(download_dest):
    """
    Imports the path in download_dest (install_idaes_workshop_materials.py module)
    """
    install_module = import_file(download_dest)
    print('... importing install module')
    return install_module

if __name__ == '__main__':
    download_dest = download_install_module()
    iwm = import_install_module(download_dest)
    iwm.execute()




##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module is a command-line script for executing convergence evaluation
testing on IDAES models.

Convergence evaluation testing is used to verify reliable convergence of a
model over a range of conditions for inputs and parameters. The developer of
the test must create a ConvergenceEvaluation class prior to executing any
convergence testing (see convergence_base.py for documentation).

Convergence evaluation testing is a two step process. In the first step, a json
file is created that contains a set of points sampled from the provided inputs.
This step only needs to be done once - up front. The second step, which should
be executed any time there is a major code change that could impact the model,
takes that set of sampled points and solves the model at each of the points,
collecting convergence statistics (success/failure, iterations, and solution
time).

To find help on convergence.py::

   $ python convergence.py --help

You will see that there are some subcommands. To find help on a particular
subcommand::

   $ python convergence.py  <subcommand> --help

To create a sample file, you can use a command-line like the following (this
should be done once by the  model developer for a few different sample sizes)::

   $ python ../../../core/util/convergence/convergence.py create-sample-file
         -s PressureChanger-10.json
         -N 10 --seed=42
         -e idaes.models.convergence.pressure_changer.
             pressure_changer_conv_eval.PressureChangerConvergenceEvaluation

More commonly, to run the convergence evaluation::

   $ python ../../../core/util/convergence/convergence.py run-eval
         -s PressureChanger-10.json

Note that the convergence evaluation can also be run in parallel if you have
installed MPI and mpi4py using a command line like the following::

   $ mpirun -np 4 python ../../../core/util/convergence/convergence.py run-eval
         -s PressureChanger-10.json

"""
import argparse
import logging
import os
import sys
#
import idaes.core.util.convergence.convergence_base as cb
from idaes.dmf import DMF
from idaes.dmf.errors import DMFError

_log = logging.getLogger(__name__)
_f = logging.Formatter(fmt='%(asctime)s %(name)s [%(levelname)s] %(message)s')
_h = logging.StreamHandler()
_h.setFormatter(_f)
_log.addHandler(_h)


def _parse_arguments():
    # register the command line arguments
    parser = argparse.ArgumentParser(prog='convergence')
    subparsers = parser.add_subparsers()
    subparsers.required = True
    subparsers.help = "Set the subcommand:"
    subparsers.metavar = 'cmd'
    subparsers.dest = 'command'
    run_report_subparser = \
        subparsers.add_parser('run-eval',
                              help="runs the convergence evaluation"
                                   " defined in the sample file given by (-s)")

    create_sample_file_subparser = \
        subparsers.add_parser('create-sample-file',
                              help="create a json file with a specified number"
                                   " of sample points (-N) based on the input"
                                   " parameter distributions defined in the"
                                   " convergence evaluation class (-e)")

    # create validation function for file existence
    def file_exists_validate(parser_obj, arg):
        if not os.path.exists(arg) or not os.path.isfile(arg):
            parser_obj.error('File: {} does not exist.'.format(arg))
        else:
            return arg

    run_report_subparser.add_argument('-s', '--sample-file',
                                      # nargs=1,
                                      default=None,
                                      type=lambda arg: file_exists_validate(
                                              parser, arg),
                                      # choices=
                                      required=True,
                                      help="Path to the json file that "
                                           "contains the points that should be"
                                           " executed in the convergence"
                                           " evaluation.",
                                      metavar='filepath',
                                      dest='sample_file')
    run_report_subparser.add_argument('-D', '--dmf', dest='dmfcfg',
                                      metavar='DIR', default=None,
                                      help='Use DMF configuration at DIR '
                                           '(default=do not use DMF)')
    run_report_subparser.add_argument('-v', '--verbose', dest='vb',
                                      action='count', default=0,
                                      help='Increase output verbosity')

    create_sample_file_subparser.add_argument(
                '-e',
                '--convergence-evaluation-class',
                # nargs=1,
                default=None,
                type=str,
                # choices=
                required=True,
                help="Specify the module.class that identifies the convergence"
                     " evaluation object that should be used to identify the"
                     " input parameters for sampling.",
                metavar='class_str',
                dest='convergence_evaluation_class_str')

    create_sample_file_subparser.add_argument(
                '-s', '--sample-file',
                # nargs=1,
                default=None,
                type=str,
                # choices=
                required=True,
                help="Path to the points file that should be created.",
                metavar='filepath',
                dest='sample_file')

    create_sample_file_subparser.add_argument(
                '-N', '--number-samples',
                # nargs=1,
                default=None,
                type=int,
                # choices=
                required=False,
                help="Used with subcommands: 'create-sample-file'. The number"
                     " of samples that should be generated when creating"
                     " the sample points file.",
                metavar='N',
                dest='number_samples')

    create_sample_file_subparser.add_argument(
                '--seed',
                # nargs=1,
                default=None,
                type=int,
                # choices=
                required=False,
                help="Used with subcommands: 'create-sample-file'. The seed"
                     " value that should be used when generating the random"
                     " samples. If None, a default seeding will be used.",
                metavar='seed',
                dest='seed')

    return parser.parse_args()


def main():
    args = _parse_arguments()

    if args.vb > 1:
        _log.setLevel(logging.DEBUG)
    elif args.vb > 0:
        _log.setLevel(logging.INFO)
    else:
        _log.setLevel(logging.WARN)

    if args.command == 'create-sample-file':
        assert args.number_samples is not None
        assert args.sample_file is not None
        assert args.convergence_evaluation_class_str is not None

        try:
            conv_eval_class = cb._class_import(
                                    args.convergence_evaluation_class_str)
            conv_eval = conv_eval_class()
        except Exception as e:
            print('Failed to find the specified convergence_evaluation_class '
                  'with error: {}'.format(str(e)))
            raise ValueError(
                    'Invalid convergence_evaluation_class specified (-e).')

        spec = conv_eval.get_specification()
        cb.write_sample_file(eval_spec=spec,
                             filename=args.sample_file,
                             convergence_evaluation_class_str=
                                 args.convergence_evaluation_class_str,
                             n_points=args.number_samples,
                             seed=args.seed)
    else:
        if args.dmfcfg is None:
            dmf = None
        else:
            try:
                dmf = DMF(args.dmfcfg)
            except DMFError as err:
                _log.error('Unable to init DMF: {}'.format(err))
                return -1
        (inputs, samples, results) = \
            cb.run_convergence_evaluation_from_sample_file(
                    sample_file=args.sample_file)
        if results is not None:
            cb.save_convergence_statistics(inputs, results, dmf=dmf)
    return 0


if __name__ == '__main__':
    sys.exit(main())
